# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Wikipedia dataset containing cleaned articles of all languages."""

import bz2
import dataclasses
import json
import re

from absl import flags  # pylint:disable=g-bad-import-order,g-import-not-at-top
from absl import logging
from etils import epath
import tensorflow_datasets.public_api as tfds

FLAGS = flags.FLAGS

_CITATION = """\
@ONLINE {wikidump,
    author = "Wikimedia Foundation",
    title  = "Wikimedia Downloads",
    url    = "https://dumps.wikimedia.org"
}
"""

_DESCRIPTION = """\
Wikipedia dataset containing cleaned articles of all languages.
The datasets are built from the Wikipedia dump
(https://dumps.wikimedia.org/) with one split per language. Each example
contains the content of one full Wikipedia article with cleaning to strip
markdown and unwanted sections (references, etc.).
"""

_LICENSE = (
    "This work is licensed under the Creative Commons Attribution-ShareAlike "
    "3.0 Unported License. To view a copy of this license, visit "
    "http://creativecommons.org/licenses/by-sa/3.0/ or send a letter to "
    "Creative Commons, PO Box 1866, Mountain View, CA 94042, USA."
)

# Source: https://en.wikipedia.org/wiki/List_of_Wikipedias (accessed 3/1/2019)
# Removed because no articles: hz.
WIKIPEDIA_LANGUAGES = [
    "aa",
    "ab",
    "ace",
    "ady",
    "af",
    "ak",
    "als",
    "am",
    "an",
    "ang",
    "ar",
    "arc",
    "arz",
    "as",
    "ast",
    "atj",
    "av",
    "ay",
    "az",
    "azb",
    "ba",
    "bar",
    "bat-smg",
    "bcl",
    "be",
    "be-x-old",
    "bg",
    "bh",
    "bi",
    "bjn",
    "bm",
    "bn",
    "bo",
    "bpy",
    "br",
    "bs",
    "bug",
    "bxr",
    "ca",
    "cbk-zam",
    "cdo",
    "ce",
    "ceb",
    "ch",
    "cho",
    "chr",
    "chy",
    "ckb",
    "co",
    "cr",
    "crh",
    "cs",
    "csb",
    "cu",
    "cv",
    "cy",
    "da",
    "de",
    "din",
    "diq",
    "dsb",
    "dty",
    "dv",
    "dz",
    "ee",
    "el",
    "eml",
    "en",
    "eo",
    "es",
    "et",
    "eu",
    "ext",
    "fa",
    "ff",
    "fi",
    "fiu-vro",
    "fj",
    "fo",
    "fr",
    "frp",
    "frr",
    "fur",
    "fy",
    "ga",
    "gag",
    "gan",
    "gd",
    "gl",
    "glk",
    "gn",
    "gom",
    "gor",
    "got",
    "gu",
    "gv",
    "ha",
    "hak",
    "haw",
    "he",
    "hi",
    "hif",
    "ho",
    "hr",
    "hsb",
    "ht",
    "hu",
    "hy",
    "ia",
    "id",
    "ie",
    "ig",
    "ii",
    "ik",
    "ilo",
    "inh",
    "io",
    "is",
    "it",
    "iu",
    "ja",
    "jam",
    "jbo",
    "jv",
    "ka",
    "kaa",
    "kab",
    "kbd",
    "kbp",
    "kg",
    "ki",
    "kj",
    "kk",
    "kl",
    "km",
    "kn",
    "ko",
    "koi",
    "krc",
    "ks",
    "ksh",
    "ku",
    "kv",
    "kw",
    "ky",
    "la",
    "lad",
    "lb",
    "lbe",
    "lez",
    "lfn",
    "lg",
    "li",
    "lij",
    "lmo",
    "ln",
    "lo",
    "lrc",
    "lt",
    "ltg",
    "lv",
    "mai",
    "map-bms",
    "mdf",
    "mg",
    "mh",
    "mhr",
    "mi",
    "min",
    "mk",
    "ml",
    "mn",
    "mr",
    "mrj",
    "ms",
    "mt",
    "mus",
    "mwl",
    "my",
    "myv",
    "mzn",
    "na",
    "nah",
    "nap",
    "nds",
    "nds-nl",
    "ne",
    "new",
    "ng",
    "nl",
    "nn",
    "no",
    "nov",
    "nrm",
    "nso",
    "nv",
    "ny",
    "oc",
    "olo",
    "om",
    "or",
    "os",
    "pa",
    "pag",
    "pam",
    "pap",
    "pcd",
    "pdc",
    "pfl",
    "pi",
    "pih",
    "pl",
    "pms",
    "pnb",
    "pnt",
    "ps",
    "pt",
    "qu",
    "rm",
    "rmy",
    "rn",
    "ro",
    "roa-rup",
    "roa-tara",
    "ru",
    "rue",
    "rw",
    "sa",
    "sah",
    "sat",
    "sc",
    "scn",
    "sco",
    "sd",
    "se",
    "sg",
    "sh",
    "si",
    "simple",
    "sk",
    "sl",
    "sm",
    "sn",
    "so",
    "sq",
    "sr",
    "srn",
    "ss",
    "st",
    "stq",
    "su",
    "sv",
    "sw",
    "szl",
    "ta",
    "tcy",
    "te",
    "tet",
    "tg",
    "th",
    "ti",
    "tk",
    "tl",
    "tn",
    "to",
    "tpi",
    "tr",
    "ts",
    "tt",
    "tum",
    "tw",
    "ty",
    "tyv",
    "udm",
    "ug",
    "uk",
    "ur",
    "uz",
    "ve",
    "vec",
    "vep",
    "vi",
    "vls",
    "vo",
    "wa",
    "war",
    "wo",
    "wuu",
    "xal",
    "xh",
    "xmf",
    "yi",
    "yo",
    "za",
    "zea",
    "zh",
    "zh-classical",
    "zh-min-nan",
    "zh-yue",
    "zu",
]


_EMPTY_LANGUAGES = {"20230601": {"aa", "na"}}


# Use mirror to avoid download caps.
_BASE_URL_TMPL = (
    "https://mirror.accum.se/mirror/wikimedia.org/dumps/{lang}wiki/{date}/"
)
_BASE_URL_TMPL_OLD = "https://dumps.wikimedia.your.org/{lang}wiki/{date}/"
_INFO_FILE = "dumpstatus.json"


class WikipediaConfig(tfds.core.BuilderConfig):
  """BuilderConfig for Wikipedia."""

  def __init__(self, *, language=None, date=None, **kwargs):
    """BuilderConfig for Wikipedia.

    Args:
      language: string, the language code for the Wikipedia dump to use.
      date: string, date of the Wikipedia dump in YYYYMMDD format. A list of
        available dates can be found at https://dumps.wikimedia.org/enwiki/.
      **kwargs: keyword arguments forwarded to super.
    """
    super(WikipediaConfig, self).__init__(
        name=f"{date}.{language}",
        description=(
            f"Wikipedia dataset for {language}, parsed from {date} dump."
        ),
        **kwargs,
    )
    self.date = date
    self.language = language


@dataclasses.dataclass(frozen=True)
class WikiPage:
  id: int
  title: str
  text: str


def _builder_configs_for(snapshot: str):
  # We put English as the first language so that it's the default.
  configs = [WikipediaConfig(language="en", date=snapshot)]
  for lang in WIKIPEDIA_LANGUAGES:
    if lang == "en" or "-" in lang:
      continue
    if snapshot in _EMPTY_LANGUAGES and lang in _EMPTY_LANGUAGES[snapshot]:
      continue
    configs.append(WikipediaConfig(language=lang, date=snapshot))
  return configs


class Wikipedia(tfds.core.BeamBasedBuilder):
  """Wikipedia dataset."""

  VERSION = tfds.core.Version("1.0.0")
  RELEASE_NOTES = {
      "1.0.0": "New split API (https://tensorflow.org/datasets/splits)",
  }

  BUILDER_CONFIGS = (
      _builder_configs_for("20230601")
      # Old versions files do not exists anymore but config are kept as
      # previously generated datasets can still be read.
      + _builder_configs_for("20230201")
      + _builder_configs_for("20220620")
      + _builder_configs_for("20201201")
      + _builder_configs_for("20200301")
      + _builder_configs_for("20190301")
  )

  def _info(self):
    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        features=tfds.features.FeaturesDict({
            "title": tfds.features.Text(),
            "text": tfds.features.Text(),
        }),
        # No default supervised_keys.
        supervised_keys=None,
        homepage="https://dumps.wikimedia.org",
        citation=_CITATION,
        license=_LICENSE,
    )

  def _split_generators(self, dl_manager):
    def _base_url(lang):
      tmpl = _BASE_URL_TMPL
      if self.builder_config.date <= "20220620":
        tmpl = _BASE_URL_TMPL_OLD
      return tmpl.format(
          lang=lang.replace("-", "_"), date=self._builder_config.date
      )

    lang = self._builder_config.language

    info_url = _base_url(lang) + _INFO_FILE
    # Use dictionary since testing mock always returns the same result.
    downloaded_files = dl_manager.download_and_extract({"info": info_url})

    xml_urls = []
    total_bytes = 0
    info_path = epath.Path(downloaded_files["info"])
    dump_info = json.loads(info_path.read_text("utf-8"))
    multistream_dump_info = dump_info["jobs"]["articlesmultistreamdump"]
    assert (
        multistream_dump_info["status"] == "done"
    ), "Specified dump (%s) multistream status is not 'done': %s" % (
        _base_url(lang),
        multistream_dump_info["status"],
    )

    for fname, info in multistream_dump_info["files"].items():
      if ".xml" not in fname:
        continue
      total_bytes += info["size"]
      xml_urls.append(_base_url(lang) + fname)

    # Use dictionary since testing mock always returns the same result.
    downloaded_files = dl_manager.download({"xml": xml_urls})

    return {
        tfds.Split.TRAIN: self._generate_examples(downloaded_files["xml"], lang)
    }

  def _generate_examples(self, filepaths, language):
    """Build PCollection of examples in the raw (text) form."""
    beam = tfds.core.lazy_imports.apache_beam
    mwparserfromhell = tfds.core.lazy_imports.mwparserfromhell
    mwxml = tfds.core.lazy_imports.mwxml

    def _extract_content(filepath):
      """Extracts article content from a single WikiMedia XML file."""
      filepath = epath.Path(filepath)
      logging.info("generating examples from = %s", filepath)
      with filepath.open("rb") as f:
        f = bz2.BZ2File(filename=f)
        dump = mwxml.Dump.from_file(f)
        for page in dump:
          beam.metrics.Metrics.counter(language, "pages-read").inc()
          if page is None:
            continue
          # Filter pages that are not in the "main" namespace.
          if page.namespace != 0:
            continue
          beam.metrics.Metrics.counter(language, "pages-in-main-read").inc()

          try:
            revision = next(iter(page))  # A single revision in dumps.
          except (TypeError, StopIteration):
            beam.metrics.Metrics.counter(language, "no-revision-error").inc()
            continue

          if not revision.text:
            beam.metrics.Metrics.counter(language, "empty-revision-error").inc()
            continue

          if revision.text[:9].lower().startswith("#redirect"):
            beam.metrics.Metrics.counter(language, "filtered-redirects").inc()
            continue

          yield WikiPage(id=page.id, title=page.title, text=revision.text)

    # Filters for references, tables, and file/image links.
    re_rm_wikilink = re.compile(
        "^(?:File|Image|Media):", flags=re.IGNORECASE | re.UNICODE
    )

    def _parse_and_clean_wikicode(raw_content):
      """Strips formatting and unwanted sections from raw page content."""
      wikicode = mwparserfromhell.parse(raw_content)

      def rm_wikilink(obj):
        return bool(re_rm_wikilink.match(str(obj.title)))  # pytype: disable=wrong-arg-types

      def rm_tag(obj):
        return str(obj.tag) in {"ref", "table"}

      def rm_template(obj):
        return obj.name.lower() in {
            "reflist",
            "notelist",
            "notelist-ua",
            "notelist-lr",
            "notelist-ur",
            "notelist-lg",
        }

      def try_remove_obj(obj, section):
        try:
          section.remove(obj)
        except ValueError:
          # For unknown reasons, objects are sometimes not found.
          pass

      section_text = []
      # Filter individual sections to clean.
      for section in wikicode.get_sections(
          flat=True, include_lead=True, include_headings=True
      ):
        for obj in section.ifilter_wikilinks(
            matches=rm_wikilink, recursive=True
        ):
          try_remove_obj(obj, section)
        for obj in section.ifilter_templates(
            matches=rm_template, recursive=True
        ):
          try_remove_obj(obj, section)
        for obj in section.ifilter_tags(matches=rm_tag, recursive=True):
          try_remove_obj(obj, section)

        section_text.append(section.strip_code().strip())
      return "\n\n".join(section_text)

    def _parse_page(page: WikiPage):
      beam.metrics.Metrics.counter(language, "extracted-examples").inc()
      try:
        text = _parse_and_clean_wikicode(page.text)
      except tfds.core.lazy_imports.mwparserfromhell.parser.ParserError as e:
        beam.metrics.Metrics.counter(language, "parser-error").inc()
        logging.error("mwparserfromhell ParseError: %s", e)
        return

      if not text:
        beam.metrics.Metrics.counter(language, "empty-clean-examples").inc()
        return

      beam.metrics.Metrics.counter(language, "cleaned-examples").inc()

      yield page.id, {
          "title": page.title.replace("_", " "),
          "text": text,
      }

    return (
        beam.Create(filepaths)
        | beam.FlatMap(_extract_content)
        | beam.Reshuffle()
        | beam.FlatMap(_parse_page)
    )



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""API to define datasets."""

# Allow to use `tfds.core.Path` in dataset implementation which seems more
# natural than having to import a third party module.
from etils.epath import Path

from tensorflow_datasets.core import community
from tensorflow_datasets.core.dataset_builder import BeamBasedBuilder
from tensorflow_datasets.core.dataset_builder import BuilderConfig
from tensorflow_datasets.core.dataset_builder import DatasetBuilder
from tensorflow_datasets.core.dataset_builder import GeneratorBasedBuilder

from tensorflow_datasets.core.dataset_info import BeamMetadataDict
from tensorflow_datasets.core.dataset_info import DatasetIdentity
from tensorflow_datasets.core.dataset_info import DatasetInfo
from tensorflow_datasets.core.dataset_info import Metadata
from tensorflow_datasets.core.dataset_info import MetadataDict

from tensorflow_datasets.core.example_serializer import ExampleSerializer

from tensorflow_datasets.core.file_adapters import FileFormat

from tensorflow_datasets.core.lazy_imports_lib import lazy_imports

from tensorflow_datasets.core.load import DatasetCollectionLoader

from tensorflow_datasets.core.naming import ShardedFileTemplate

from tensorflow_datasets.core.registered import DatasetNotFoundError

from tensorflow_datasets.core.sequential_writer import SequentialWriter

from tensorflow_datasets.core.split_builder import SplitGeneratorLegacy as SplitGenerator

from tensorflow_datasets.core.splits import ReadInstruction
from tensorflow_datasets.core.splits import Split
from tensorflow_datasets.core.splits import SplitDict
from tensorflow_datasets.core.splits import SplitInfo
from tensorflow_datasets.core.splits import SubSplitInfo

from tensorflow_datasets.core.utils import Experiment
from tensorflow_datasets.core.utils import gcs_path
from tensorflow_datasets.core.utils import lazy_imports_utils
from tensorflow_datasets.core.utils import tfds_path
from tensorflow_datasets.core.utils import Version
from tensorflow_datasets.core.utils.benchmark import BenchmarkResult
from tensorflow_datasets.core.utils.file_utils import add_data_dir
from tensorflow_datasets.core.utils.file_utils import as_path


def benchmark(*args, **kwargs):
  raise DeprecationWarning(
      "`tfds.core.benchmark` has been renamed to `tfds.benchmark`"
  )


__all__ = [
    "add_data_dir",
    "as_path",
    "BenchmarkResult",
    "BeamBasedBuilder",
    "BeamMetadataDict",
    "BuilderConfig",
    "DatasetBuilder",
    "DatasetCollectionLoader",
    "DatasetInfo",
    "DatasetIdentity",
    "DatasetNotFoundError",
    "Experiment",
    "FileFormat",
    "GeneratorBasedBuilder",
    "gcs_path",
    "lazy_imports",
    "Metadata",
    "MetadataDict",
    "Path",
    "ReadInstruction",
    "SequentialWriter",
    "ShardedFileTemplate",
    "SplitDict",
    "SplitGenerator",
    "SplitInfo",
    "tfds_path",
    "Version",
]



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Async download API with checksum verification. No business logic."""

import concurrent.futures
import contextlib
import dataclasses
import functools
import hashlib
import io
import os
import re
from typing import Any, ContextManager, Iterable, Iterator, Optional, Tuple, Union
import urllib

from etils import epath
import promise
import requests
from tensorflow_datasets.core import units
from tensorflow_datasets.core import utils
from tensorflow_datasets.core.download import checksums as checksums_lib
from tensorflow_datasets.core.utils import tqdm_utils
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf

_DRIVE_URL = re.compile(r'^https://drive\.google\.com/')
MAX_RETRIES = 10

# Response interface. Has `.url` and `.headers` attribute
Response = Union[requests.Response, urllib.response.addinfourl]


@dataclasses.dataclass(eq=False, frozen=True)
class DownloadResult:
  path: epath.Path
  url_info: checksums_lib.UrlInfo


@utils.memoize()
def get_downloader(*args: Any, **kwargs: Any) -> '_Downloader':
  return _Downloader(*args, **kwargs)


def _filename_from_content_disposition(
    content_disposition: str,
) -> Optional[str]:
  """Extract the filename from the content disposition.

  Parse the content_definition as defined in:
  https://tools.ietf.org/html/rfc2616

  Note:

   * If both encoded (`filename*=`) and ascii (filename=) name are defined,
     the function returns the ascii name, as encoding might create issue on
     some systems
   * If only the encoded name is defined (e.g.
     `filename*=UTF-8''%e2%82%ac.txt`), the function return None as this is
     not yet supported.

  Args:
      content_disposition: String to parse.

  Returns:
      filename: The filename, or None if filename could not be parsed
  """
  match = re.findall(
      # Regex (see unittests for examples):
      # ` *` : Strip eventual whitespaces
      # `['"]?` : Filename is optionally wrapped in quote
      # `([^;\r\n"']+)` : Filename can be any symbol except those
      # `;?` : Stop when encountering optional `;`
      r"""filename= *['"]?([^;\r\n"']+)['"]? *;?""",
      content_disposition,
      flags=re.IGNORECASE,
  )
  if not match:
    return None
  elif len(match) != 1:
    raise ValueError(
        f'Error while parsing filename for: {content_disposition}\n'
        f'Multiple filename detected: {list(match)}'
    )
  return os.path.basename(match[0].rstrip())


def _get_filename(response: Response) -> str:
  content_disposition = response.headers.get('content-disposition', None)
  if content_disposition:
    filename = _filename_from_content_disposition(content_disposition)
    if filename:
      return filename
  # Otherwise, fallback on extracting the name from the url.
  return utils.basename_from_url(response.url)


class DownloadError(Exception):
  pass


class _Downloader(object):
  """Class providing async download API with checksum validation.

  Do not instantiate this class directly. Instead, call `get_downloader()`.
  """

  _DEFAULT_MAX_SIMULTANEOUS_DOWNLOADS = 50
  _pbar_url: tqdm_utils._TqdmPbarAsync
  _pbar_dl_size: tqdm_utils._TqdmPbarAsync

  def __init__(
      self,
      max_simultaneous_downloads: Optional[
          int
      ] = _DEFAULT_MAX_SIMULTANEOUS_DOWNLOADS,
      checksumer=None,
  ):
    """Init _Downloader instance.

    Args:
      max_simultaneous_downloads: `int`, optional max number of simultaneous
        downloads. If None then it defaults to
        `self._DEFAULT_MAX_SIMULTANEOUS_DOWNLOADS`.
      checksumer: `hashlib.HASH`. Defaults to `hashlib.sha256`.
    """
    self._executor = concurrent.futures.ThreadPoolExecutor(
        max_workers=max_simultaneous_downloads
        or self._DEFAULT_MAX_SIMULTANEOUS_DOWNLOADS
    )
    self._checksumer_cls = checksumer or hashlib.sha256

  @contextlib.contextmanager
  def tqdm(self) -> Iterator[None]:
    """Add a progression bar for the current download."""
    async_tqdm = utils.async_tqdm
    with async_tqdm(total=0, desc='Dl Completed...', unit=' url') as pbar_url:
      with async_tqdm(total=0, desc='Dl Size...', unit=' MiB') as pbar_dl_size:
        self._pbar_url = pbar_url
        self._pbar_dl_size = pbar_dl_size
        yield

  def increase_tqdm(self, dl_result: DownloadResult) -> None:
    """Update the tqdm bars to visually indicate the dl_result is downloaded."""
    self._pbar_url.update_total(1)
    self._pbar_url.update(1)
    if dl_result.url_info:  # Info unknown for manually downloaded files
      self._pbar_dl_size.update_total(dl_result.url_info.size)
      self._pbar_dl_size.update(dl_result.url_info.size)

  def download(
      self, url: str, destination_path: str, verify: bool = True
  ) -> 'promise.Promise[concurrent.futures.Future[DownloadResult]]':
    """Download url to given path.

    Returns Promise -> sha256 of downloaded file.

    Args:
      url: address of resource to download.
      destination_path: `str`, path to directory where to download the resource.
      verify: whether to verify ssl certificates

    Returns:
      Promise obj -> (`str`, int): (downloaded object checksum, size in bytes).
    """
    destination_path = os.fspath(destination_path)
    self._pbar_url.update_total(1)
    future = self._executor.submit(
        self._sync_download, url, destination_path, verify
    )
    return promise.Promise.resolve(future)

  def _sync_file_copy(
      self,
      filepath: str,
      destination_path: str,
  ) -> DownloadResult:
    """Downloads the file through `tf.io.gfile` API."""
    filename = os.path.basename(filepath)
    out_path = os.path.join(destination_path, filename)
    tf.io.gfile.copy(filepath, out_path)
    url_info = checksums_lib.compute_url_info(
        out_path, checksum_cls=self._checksumer_cls
    )
    self._pbar_dl_size.update_total(url_info.size)
    self._pbar_dl_size.update(url_info.size)
    self._pbar_url.update(1)
    return DownloadResult(path=epath.Path(out_path), url_info=url_info)

  def _sync_download(
      self, url: str, destination_path: str, verify: bool = True
  ) -> DownloadResult:
    """Synchronous version of `download` method.

    To download through a proxy, the `HTTP_PROXY`, `HTTPS_PROXY`,
    `REQUESTS_CA_BUNDLE`,... environment variables can be exported, as
    described in:
    https://requests.readthedocs.io/en/master/user/advanced/#proxies

    Args:
      url: url to download
      destination_path: path where to write it
      verify: whether to verify ssl certificates

    Returns:
      None

    Raises:
      DownloadError: when download fails.
    """
    try:
      # If url is on a filesystem that gfile understands, use copy. Otherwise,
      # use requests (http) or urllib (ftp).
      if not url.startswith('http'):
        return self._sync_file_copy(url, destination_path)
    except tf.errors.UnimplementedError:
      pass

    with _open_url(url, verify=verify) as (response, iter_content):
      fname = _get_filename(response)
      path = os.path.join(destination_path, fname)
      size = 0

      # Initialize the download size progress bar
      size_mb = 0
      unit_mb = units.MiB
      total_size = int(response.headers.get('Content-length', 0)) // unit_mb
      self._pbar_dl_size.update_total(total_size)
      with tf.io.gfile.GFile(path, 'wb') as file_:
        checksum = self._checksumer_cls()
        for block in iter_content:
          size += len(block)
          checksum.update(block)
          file_.write(block)

          # Update the download size progress bar
          size_mb += len(block)
          if size_mb > unit_mb:
            self._pbar_dl_size.update(size_mb // unit_mb)
            size_mb %= unit_mb
    self._pbar_url.update(1)
    return DownloadResult(
        path=epath.Path(path),
        url_info=checksums_lib.UrlInfo(
            checksum=checksum.hexdigest(),
            size=utils.Size(size),
            filename=fname,
        ),
    )


def _open_url(
    url: str,
    **kwargs: Any,
) -> ContextManager[Tuple[Response, Iterable[bytes]]]:
  """Context manager to open an url.

  Args:
    url: The url to open
    **kwargs: Additional kwargs to forward to `request.get`.

  Returns:
    response: The url response with `.url` and `.header` attributes.
    iter_content: A `bytes` iterator which yield the content.
  """
  # Download FTP urls with `urllib`, otherwise use `requests`
  open_fn = _open_with_urllib if url.startswith('ftp') else _open_with_requests
  return open_fn(url, **kwargs)


@contextlib.contextmanager
def _open_with_requests(
    url: str,
    **kwargs: Any,
) -> Iterator[Tuple[Response, Iterable[bytes]]]:
  """Open url with request."""
  with requests.Session() as session:
    retries = requests.packages.urllib3.util.retry.Retry(
        total=MAX_RETRIES,
        backoff_factor=0.2,
        status_forcelist=[500, 502, 503, 504],
        raise_on_redirect=True,
        raise_on_status=True,
    )
    session.mount('http://', requests.adapters.HTTPAdapter(max_retries=retries))
    session.mount(
        'https://', requests.adapters.HTTPAdapter(max_retries=retries)
    )
    if _DRIVE_URL.match(url):
      url = _normalize_drive_url(url)
    with session.get(url, stream=True, **kwargs) as response:
      _assert_status(response)
      yield (response, response.iter_content(chunk_size=io.DEFAULT_BUFFER_SIZE))


@contextlib.contextmanager
def _open_with_urllib(
    url: str,
    **kwargs: Any,
) -> Iterator[Tuple[Response, Iterable[bytes]]]:
  del kwargs
  with urllib.request.urlopen(url) as response:  # pytype: disable=attribute-error
    yield (
        response,
        iter(functools.partial(response.read, io.DEFAULT_BUFFER_SIZE), b''),
    )


def _normalize_drive_url(url: str) -> str:
  """Returns Google Drive url with confirmation token."""
  # This bypasses the "Google Drive can't scan this file for viruses" warning
  # when dowloading large files.
  return url + '&confirm=t'


def _assert_status(response: requests.Response) -> None:
  """Ensure the URL response is 200."""
  if response.status_code != 200:
    raise DownloadError(
        'Failed to get url {}. HTTP code: {}.'.format(
            response.url, response.status_code
        )
    )



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""hellaswag dataset."""

import tensorflow_datasets.public_api as tfds
from tensorflow_datasets.text import hellaswag


class HellaswagTest(tfds.testing.DatasetBuilderTestCase):
  DATASET_CLASS = hellaswag.Hellaswag
  SPLITS = {
      'train': 2,
      'validation': 4,
      'test': 4,
      'train_activitynet': 1,
      'train_wikihow': 1,
      'validation_ind_activitynet': 1,
      'validation_ood_activitynet': 1,
      'test_ind_activitynet': 1,
      'test_ood_activitynet': 1,
      'validation_ind_wikihow': 1,
      'validation_ood_wikihow': 1,
      'test_ind_wikihow': 1,
      'test_ood_wikihow': 1,
  }
  DL_EXTRACT_RESULT = {
      'train': 'hellaswag_train.jsonl',
      'test': 'hellaswag_test.jsonl',
      'validation': 'hellaswag_val.jsonl',
  }
  OVERLAPPING_SPLITS = [
      'train_activitynet',
      'train_wikihow',
      'validation_ind_activitynet',
      'validation_ood_activitynet',
      'test_ind_activitynet',
      'test_ood_activitynet',
      'validation_ind_wikihow',
      'validation_ood_wikihow',
      'test_ind_wikihow',
      'test_ood_wikihow',
  ]


if __name__ == '__main__':
  tfds.testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""LAION-400M image dataset."""

import tensorflow_datasets.public_api as tfds
from tensorflow_datasets.vision_language.laion400m import laion400m


class Laion400MImagesTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for LAION-400M image dataset."""

  BUILDER_CONFIG_NAMES_TO_TEST = [laion400m.LAION400M_IMAGES_CONFIG.name]
  DATASET_CLASS = laion400m.Laion400m
  SPLITS = {
      'train': 3,
  }

  @classmethod
  def setUpClass(cls):
    laion400m.LAION400M_IMAGES_CONFIG.num_shards = 1
    super().setUpClass()


class Laion400MEmbeddingsTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for LAION-400M image dataset."""

  BUILDER_CONFIG_NAMES_TO_TEST = [laion400m.LAION400M_EMBEDDINGS_CONFIG.name]
  DATASET_CLASS = laion400m.Laion400m
  SPLITS = {
      'train': 3,
  }
  DL_EXTRACT_RESULT = {
      'img_emb_0.npy': 'img_emb_0.npy',
      'text_emb_0.npy': 'text_emb_0.npy',
      'metadata_0.parquet': 'metadata_0.parquet',
  }

  @classmethod
  def setUpClass(cls):
    laion400m.LAION400M_EMBEDDINGS_CONFIG.num_shards = 1
    super().setUpClass()


if __name__ == '__main__':
  tfds.testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for tensorflow_datasets.core.test_utils."""

import pathlib

import pytest

import tensorflow as tf

from tensorflow_datasets.testing import test_case
from tensorflow_datasets.testing import test_utils


class RunInGraphAndEagerTest(test_case.TestCase):

  def test_run_in_graph_and_eager_modes(self):
    l = []

    def inc(self, with_brackets):
      del self  # self argument is required by run_in_graph_and_eager_modes.
      mode = 'eager' if tf.executing_eagerly() else 'graph'
      with_brackets = 'with_brackets' if with_brackets else 'without_brackets'
      l.append((with_brackets, mode))

    f = test_utils.run_in_graph_and_eager_modes(inc)
    f(self, with_brackets=False)
    f = test_utils.run_in_graph_and_eager_modes()(inc)
    f(self, with_brackets=True)

    self.assertEqual(len(l), 4)
    self.assertEqual(
        set(l),
        {
            ('with_brackets', 'graph'),
            ('with_brackets', 'eager'),
            ('without_brackets', 'graph'),
            ('without_brackets', 'eager'),
        },
    )

  def test_run_in_graph_and_eager_modes_setup_in_same_mode(self):
    modes = []
    mode_name = lambda: 'eager' if tf.executing_eagerly() else 'graph'

    class ExampleTest(test_case.TestCase):

      def runTest(self):
        pass

      def setUp(self):
        super(ExampleTest, self).setUp()
        modes.append('setup_' + mode_name())

      def subTest(self, msg, **params):
        modes.append('subtest_' + msg)
        return super().subTest(msg, **params)

      @test_utils.run_in_graph_and_eager_modes
      def testBody(self):
        modes.append('run_' + mode_name())

    e = ExampleTest()
    e.setUp()
    e.testBody()

    self.assertEqual(
        modes,
        [
            'setup_eager',
            'subtest_eager_mode',
            'run_eager',
            'subtest_graph_mode',
            'setup_graph',
            'run_graph',
        ],
    )

  def test_mock_tf(self):
    # pylint: disable=g-import-not-at-top,reimported
    import tensorflow as tf_lib1
    import tensorflow as tf_lib2

    # pylint: enable=g-import-not-at-top,reimported

    def f():
      pass

    original_exists = tf_lib1.io.gfile.exists
    original_list_dir = tf_lib1.io.gfile.listdir

    self.assertIs(tf_lib1.io.gfile.exists, original_exists)
    self.assertIs(tf_lib2.io.gfile.listdir, original_list_dir)

    with test_utils.mock_tf('tf.io.gfile', exists=f):
      # Both aliases should have been patched
      self.assertIs(tf_lib1.io.gfile.exists, f)
      self.assertIs(tf_lib2.io.gfile.exists, f)
      self.assertIs(tf_lib1.io.gfile.listdir, original_list_dir)
      self.assertIs(tf_lib2.io.gfile.listdir, original_list_dir)

    self.assertIsNot(tf_lib1.io.gfile.exists, f)
    self.assertIsNot(tf_lib2.io.gfile.exists, f)
    self.assertIs(tf_lib1.io.gfile.exists, original_exists)
    self.assertIs(tf_lib1.io.gfile.listdir, original_list_dir)
    self.assertIs(tf_lib2.io.gfile.exists, original_exists)
    self.assertIs(tf_lib2.io.gfile.listdir, original_list_dir)

    with test_utils.mock_tf('tf.io.gfile.exists', f):
      self.assertIs(tf_lib1.io.gfile.exists, f)
      self.assertIs(tf_lib2.io.gfile.exists, f)
      self.assertIs(tf_lib1.io.gfile.listdir, original_list_dir)
      self.assertIs(tf_lib2.io.gfile.listdir, original_list_dir)


@pytest.mark.parametrize(
    'as_path_fn', [pathlib.Path, str]  # Test both PathLike and str
)
def test_mock_fs(as_path_fn):
  _p = as_path_fn  # pylint: disable=invalid-name
  fs = test_utils.MockFs()
  with fs.mock():
    fs.add_file(_p('/path/to/file1'), 'Content of file 1')
    fs.add_file(_p('/path/file.txt'), 'Content of file.txt')

    # Test `tf.io.gfile.exists`
    assert tf.io.gfile.exists(_p('/path/to/file1'))
    assert tf.io.gfile.exists(_p('/path/to/'))
    assert tf.io.gfile.exists(_p('/path/to'))
    assert tf.io.gfile.exists(_p('/path'))
    assert not tf.io.gfile.exists(_p('/pat'))
    assert not tf.io.gfile.exists(_p('/path/to/file1_nonexisting'))
    assert not tf.io.gfile.exists(_p('/path/to_file1_nonexisting'))

    # Test `tf.io.gfile.exists` (relative path)
    fs.add_file(_p('relative_path/to/file.txt'), 'Content')
    assert tf.io.gfile.exists(_p('relative_path/to/file.txt'))
    assert tf.io.gfile.exists(_p('relative_path/to/'))
    assert tf.io.gfile.exists(_p('relative_path/to'))
    assert tf.io.gfile.exists(_p('relative_path'))

    # Test `tf.io.gfile.GFile` (write and read mode)
    with tf.io.gfile.GFile(_p('/path/to/file2'), 'w') as f:
      f.write('Content of file 2 (old)')
    assert fs.read_file('/path/to/file2') == 'Content of file 2 (old)'
    with tf.io.gfile.GFile(_p('/path/to/file2'), 'w') as f:
      f.write('Content of file 2 (new)')
    assert fs.read_file('/path/to/file2') == 'Content of file 2 (new)'
    with tf.io.gfile.GFile(_p('/path/to/file2'), 'r') as f:
      assert f.read() == 'Content of file 2 (new)'

    # Test `tf.io.gfile.rename`
    assert fs.read_file('/path/to/file1') == 'Content of file 1'
    tf.io.gfile.rename(_p('/path/to/file1'), _p('/path/to/file1_moved'))
    assert not tf.io.gfile.exists('/path/to/file1')
    assert fs.read_file('/path/to/file1_moved') == 'Content of file 1'

    # Test `tf.io.gfile.listdir`
    assert set(tf.io.gfile.listdir(_p('/path/to'))) == set(
        tf.io.gfile.listdir(_p('/path/to/'))
    )
    assert set(tf.io.gfile.listdir(_p('/path/to'))) == {'file1_moved', 'file2'}
    assert set(tf.io.gfile.listdir(_p('/path'))) == {'file.txt', 'to'}

    # Test `MockFs.files`
    assert fs.read_file('/path/to/file2') == 'Content of file 2 (new)'
    assert fs.read_file('/path/to/file1_moved') == 'Content of file 1'
    assert fs.read_file('/path/file.txt') == 'Content of file.txt'
    assert fs.read_file('relative_path/to/file.txt') == 'Content'


def test_mock_fs_gcs():
  with test_utils.MockFs():
    tf.io.gfile.makedirs('gs://bucket/dir/subdir')
    assert tf.io.gfile.exists('gs://bucket/dir')
    assert tf.io.gfile.exists('gs://bucket/dir/subdir')

    with tf.io.gfile.GFile('gs://bucket/dir/file.txt', 'w') as f:
      f.write('Content of file')

    assert set(tf.io.gfile.listdir('gs://bucket/dir')) == {
        'file.txt',
        'subdir',
    }

    bs = 'big' + 'store'

    # Test
    # * gs
    # * absolute path
    # * relative path
    assert tf.io.gfile.glob('gs://bucket/*/file.txt') == [
        'gs://bucket/dir/file.txt',
    ]
    assert tf.io.gfile.glob(f'/{bs}/bucket/*/file.txt') == [
        f'/{bs}/bucket/dir/file.txt',
    ]
    assert tf.io.gfile.glob(f'{bs}/bucket/*/file.txt') == [
        f'{bs}/bucket/dir/file.txt',
    ]



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""conll2002 dataset."""
import collections

from tensorflow_datasets.core.dataset_builders.conll import conll_dataset_builder
from tensorflow_datasets.core.dataset_builders.conll import conll_dataset_builder_utils as conll_lib
from tensorflow_datasets.core.features.class_label_feature import ClassLabel
from tensorflow_datasets.core.features.sequence_feature import Sequence
from tensorflow_datasets.core.features.text_feature import Text
import tensorflow_datasets.public_api as tfds

_DESCRIPTION = """
The shared task of CoNLL-2002 concerns language-independent named entity
recognition. The types of named entities include: persons, locations,
organizations and names of miscellaneous entities that do not belong to the
previous three groups. The participants of the shared task were offered training
and test data for at least two languages. Information sources other than the
training data might have been used in this shared task.
"""

_CITATION = """
@inproceedings{tjong-kim-sang-2002-introduction,
    title = "Introduction to the {C}o{NLL}-2002 Shared Task: Language-Independent Named Entity Recognition",
    author = "Tjong Kim Sang, Erik F.",
    booktitle = "{COLING}-02: The 6th Conference on Natural Language Learning 2002 ({C}o{NLL}-2002)",
    year = "2002",
    url = "https://aclanthology.org/W02-2024",
}
"""

_URL = "https://raw.githubusercontent.com/teropa/nlp/master/resources/corpora/conll2002/"

ES_POS_TAGS = (
    "AO",
    "AQ",
    "CC",
    "CS",
    "DA",
    "DE",
    "DD",
    "DI",
    "DN",
    "DP",
    "DT",
    "Faa",
    "Fat",
    "Fc",
    "Fd",
    "Fe",
    "Fg",
    "Fh",
    "Fia",
    "Fit",
    "Fp",
    "Fpa",
    "Fpt",
    "Fs",
    "Ft",
    "Fx",
    "Fz",
    "I",
    "NC",
    "NP",
    "P0",
    "PD",
    "PI",
    "PN",
    "PP",
    "PR",
    "PT",
    "PX",
    "RG",
    "RN",
    "SP",
    "VAI",
    "VAM",
    "VAN",
    "VAP",
    "VAS",
    "VMG",
    "VMI",
    "VMM",
    "VMN",
    "VMP",
    "VMS",
    "VSG",
    "VSI",
    "VSM",
    "VSN",
    "VSP",
    "VSS",
    "Y",
    "Z",
)

NL_POS_TAGS = (
    "Adj",
    "Adv",
    "Art",
    "Conj",
    "Int",
    "Misc",
    "N",
    "Num",
    "Prep",
    "Pron",
    "Punc",
    "V",
)

ES_CONFIG = conll_dataset_builder.ConllBuilderConfig(
    name="es",
    separator=conll_lib.CONLL_2002_SEPARATOR,
    ordered_features=collections.OrderedDict({
        "tokens": Sequence(Text()),
        "pos": Sequence(ClassLabel(names=ES_POS_TAGS)),
        "ner": Sequence(ClassLabel(names=conll_lib.NER_TAGS)),
    }),
)

NL_CONFIG = conll_dataset_builder.ConllBuilderConfig(
    name="nl",
    separator=conll_lib.CONLL_2002_SEPARATOR,
    ordered_features=collections.OrderedDict({
        "tokens": Sequence(Text()),
        "pos": Sequence(ClassLabel(names=NL_POS_TAGS)),
        "ner": Sequence(ClassLabel(names=conll_lib.NER_TAGS)),
    }),
)


class Conll2002(tfds.dataset_builders.ConllDatasetBuilder):
  """DatasetBuilder for conll2002 dataset."""

  VERSION = tfds.core.Version("1.0.0")
  RELEASE_NOTES = {
      "1.0.0": "Initial release.",
  }
  BUILDER_CONFIGS = [ES_CONFIG, NL_CONFIG]

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    return self.create_dataset_info(
        description=_DESCRIPTION,
        homepage="https://aclanthology.org/W02-2024/",
        citation=_CITATION,
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""
    file_name = "esp" if self.builder_config.name == ES_CONFIG.name else "ned"
    urls_to_download = {
        "train": f"{_URL}{file_name}.train",
        "dev": f"{_URL}{file_name}.testa",
        "test": f"{_URL}{file_name}.testb",
    }

    dl_paths = dl_manager.download(urls_to_download)

    return {
        "train": self._generate_examples(dl_paths["train"]),
        "dev": self._generate_examples(dl_paths["dev"]),
        "test": self._generate_examples(dl_paths["test"]),
    }



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""CMATERdb dataset."""

import numpy as np
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

# CMATERdb constants
_CMATERDB_IMAGE_SIZE = 32
_CMATERDB_IMAGE_SHAPE = (_CMATERDB_IMAGE_SIZE, _CMATERDB_IMAGE_SIZE, 3)
# GitHub npz mirror of https://code.google.com/archive/p/cmaterdb/
_CMATERDB_TRAINING_URL = (
    "https://raw.githubusercontent.com/prabhuomkar/CMATERdb/master/"
    "datasets/{type}-numerals/training-images.npz"
)
_CMATERDB_TESTING_URL = (
    "https://raw.githubusercontent.com/prabhuomkar/CMATERdb/master/"
    "datasets/{type}-numerals/testing-images.npz"
)

_CITATION = """\
@article{Das:2012:GAB:2161007.2161320,
  author = {Das, Nibaran and Sarkar, Ram and Basu, Subhadip and Kundu, Mahantapas 
            and Nasipuri, Mita and Basu, Dipak Kumar},
  title = {A Genetic Algorithm Based Region Sampling for Selection of Local Features 
          in Handwritten Digit Recognition Application},
  journal = {Appl. Soft Comput.},
  issue_date = {May, 2012},
  volume = {12},
  number = {5},
  month = may,
  year = {2012},
  issn = {1568-4946},
  pages = {1592--1606},
  numpages = {15},
  url = {http://dx.doi.org/10.1016/j.asoc.2011.11.030},
  doi = {10.1016/j.asoc.2011.11.030},
  acmid = {2161320},
  publisher = {Elsevier Science Publishers B. V.},
  address = {Amsterdam, The Netherlands, The Netherlands},
  keywords = {Feature selection, Genetic algorithm, N-Quality consensus, 
  Optimal local regions, Region sampling, Variable sized local regions},
}
@article{Das:2012:SFC:2240301.2240421,
  author = {Das, Nibaran and Reddy, Jagan Mohan and Sarkar, Ram and Basu, Subhadip and Kundu, 
            Mahantapas and Nasipuri, Mita and Basu, Dipak Kumar},
  title = {A Statistical-topological Feature Combination for Recognition of Handwritten Numerals},
  journal = {Appl. Soft Comput.},
  issue_date = {August, 2012},
  volume = {12},
  number = {8},
  month = aug,
  year = {2012},
  issn = {1568-4946},
  pages = {2486--2495},
  numpages = {10},
  url = {http://dx.doi.org/10.1016/j.asoc.2012.03.039},
  doi = {10.1016/j.asoc.2012.03.039},
  acmid = {2240421},
  publisher = {Elsevier Science Publishers B. V.},
  address = {Amsterdam, The Netherlands, The Netherlands},
  keywords = {Character recognition, Feature combination, MPCA, PCA, SVM, Statistical, Topological},
}
"""

_DESCRIPTION = """\
This dataset contains images of -
  Handwritten Bangla numerals - balanced dataset of total 6000 Bangla numerals (32x32 RGB coloured, 6000 images), each having 600 images per class(per digit). 
  Handwritten Devanagari numerals - balanced dataset of total 3000 Devanagari numerals (32x32 RGB coloured, 3000 images), each having 300 images per class(per digit). 
  Handwritten Telugu numerals - balanced dataset of total 3000 Telugu numerals (32x32 RGB coloured, 3000 images), each having 300 images per class(per digit). 

CMATERdb is the pattern recognition database repository created at the 'Center for Microprocessor Applications for Training Education and Research' (CMATER) research lab, Jadavpur University, India.
"""


class CmaterdbConfig(tfds.core.BuilderConfig):
  """BuilderConfig for CMATERdb Config."""


class Cmaterdb(tfds.core.GeneratorBasedBuilder):
  """CMATERdb dataset."""

  BUILDER_CONFIGS = [
      CmaterdbConfig(
          name="bangla",
          description="CMATERdb Bangla Numerals",
          version="1.0.0",
      ),
      CmaterdbConfig(
          name="devanagari",
          description="CMATERdb Devangari Numerals",
          version="1.0.0",
      ),
      CmaterdbConfig(
          name="telugu",
          description="CMATERdb Telugu Numerals",
          version="1.0.0",
      ),
  ]

  def _info(self):
    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        features=tfds.features.FeaturesDict({
            "image": tfds.features.Image(shape=_CMATERDB_IMAGE_SHAPE),
            "label": tfds.features.ClassLabel(num_classes=10),
        }),
        supervised_keys=("image", "label"),
        homepage="https://code.google.com/archive/p/cmaterdb/",
        citation=_CITATION,
    )

  def _split_generators(self, dl_manager):
    """Returns SplitGenerators."""
    # Download the CMATERdb dataset by mentioned numeral
    train_path, test_path = dl_manager.download([
        _CMATERDB_TRAINING_URL.format(type=self.builder_config.name),
        _CMATERDB_TESTING_URL.format(type=self.builder_config.name),
    ])

    # CMATERdb (mirrored) provides TRAIN and TEST splits,
    # not a VALIDATION split, so we only
    # write the TRAIN and TEST splits to disk.
    return [
        tfds.core.SplitGenerator(
            name=tfds.Split.TRAIN,
            gen_kwargs=dict(data_path=train_path),
        ),
        tfds.core.SplitGenerator(
            name=tfds.Split.TEST,
            gen_kwargs=dict(data_path=test_path),
        ),
    ]

  def _generate_examples(self, data_path):
    """Generate CMATERdb examples as dicts.

    Args:
      data_path (str): Path to the data files

    Yields:
      Generator yielding the next examples
    """
    with tf.io.gfile.GFile(data_path, mode="rb") as f:
      data = np.load(f)

    data = list(zip(data["images"], data["labels"]))
    for index, (image, label) in enumerate(data):
      yield index, {
          "image": image,
          "label": label,
      }



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for dataset_markdown_builder."""
import tensorflow as tf
import tensorflow_datasets as tfds
from tensorflow_datasets.scripts.documentation import dataset_markdown_builder


def test_feature_documentation_section():
  features = tfds.features.FeaturesDict(
      feature_dict={
          'a': tfds.features.Scalar(
              dtype=tf.int64,
              doc=tfds.features.Documentation(
                  desc='a feature', value_range='From 1 to 10'
              ),
          ),
          'b': tfds.features.Text(doc='Some text'),
          'c': tfds.features.Image(
              shape=(None, None, 1), encoding_format='png'
          ),
          'd': tf.int64,
          'e': tfds.features.Sequence(tf.float32, doc='Floats'),
          'f': tfds.features.Sequence(tfds.features.Audio(), doc='Audio'),
          'g': tfds.features.Sequence(
              {
                  'a': tf.float32,
                  'b': tfds.features.Scalar(dtype=tf.int64, doc='b'),
              },
              doc='Nested',
          ),
          'h': tfds.features.Dataset(
              {
                  'a': tf.float32,
                  'b': tfds.features.Text(doc='Nested text'),
              },
              doc='Dataset of something',
          ),
      },
      doc='My test features',
  )
  section = dataset_markdown_builder.FeatureDocumentationSection()
  assert section._format_block(features.catalog_documentation()) == '\n'.join([
      'Feature | Class | Shape | Dtype | Description | Value range',
      ':------ | :---- | :---- | :---- | :---------- | :----------',
      ' | FeaturesDict |  |  | My test features | ',
      'a | Scalar |  | int64 | a feature | From 1 to 10',
      'b | Text |  | string | Some text | ',
      'c | Image | (None, None, 1) | uint8 |  | ',
      'd | Tensor |  | int64 |  | ',
      'e | Sequence(Tensor) | (None,) | float32 | Floats | ',
      'f | Sequence(Audio) | (None, None) | int64 | Audio | ',
      'g | Sequence |  |  | Nested | ',
      'g/a | Tensor |  | float32 |  | ',
      'g/b | Scalar |  | int64 | b | ',
      'h | Dataset |  |  | Dataset of something | ',
      'h/a | Tensor |  | float32 |  | ',
      'h/b | Text |  | string | Nested text | ',
  ])


def test_feature_documentation_section_missing_value_range():
  features = tfds.features.FeaturesDict(
      feature_dict={
          'a': tfds.features.Scalar(
              dtype=tf.int64, doc=tfds.features.Documentation(desc='a feature')
          ),
          'b': tfds.features.Text(doc='Some text'),
      },
      doc='My test features',
  )
  section = dataset_markdown_builder.FeatureDocumentationSection()
  assert section._format_block(features.catalog_documentation()) == '\n'.join([
      'Feature | Class | Shape | Dtype | Description',
      ':------ | :---- | :---- | :---- | :----------',
      ' | FeaturesDict |  |  | My test features',
      'a | Scalar |  | int64 | a feature',
      'b | Text |  | string | Some text',
  ])


def test_paperswithcode_section():
  tfds_to_pwc_links = {'dummy_dataset': 'https://paperswithcode/dummy_dataset'}
  pwc_section = dataset_markdown_builder.PapersWithCodeSection(
      tfds_to_pwc_links=tfds_to_pwc_links
  )

  dummy_builder = tfds.testing.DummyDataset()
  expected_output = f"""
        <a class="button button-with-icon" href="{tfds_to_pwc_links[dummy_builder.name]}">
          Explore on Papers With Code
          <span class="material-icons icon-after" aria-hidden="true">
            north_east
          </span>
        </a>
      """

  assert pwc_section.content(builder=dummy_builder) == expected_output



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for tree_utils."""

import pytest
from tensorflow_datasets.core.utils import tree_utils


def test_tree_parallel_map():
  assert tree_utils.parallel_map(
      lambda x: x * 10, {'a': [1, 2, 3], 'b': [4, 5]}
  ) == {'a': [10, 20, 30], 'b': [40, 50]}


def test_tree_parallel_map_reraise():
  def fn(x):
    raise ValueError('Bad value')

  with pytest.raises(ValueError, match='Bad value'):
    tree_utils.parallel_map(fn, [1])



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""TensorFlow compatibility utilities."""

import functools
import os
from typing import Optional

from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf


def get_single_element(ds):
  """Calls `tf.data.Dataset.get_single_element`."""
  if hasattr(ds, "get_single_element"):  # tf 2.6 and above
    return ds.get_single_element()
  else:
    return tf.data.experimental.get_single_element(ds)


def _get_option_deterministic_field() -> str:
  if hasattr(tf.data.Options(), "deterministic"):
    return "deterministic"
  return "experimental_deterministic"


def get_option_deterministic(options) -> Optional[bool]:
  """Returns the option whether the output should be in deterministic order."""
  return getattr(options, _get_option_deterministic_field())


def set_option_deterministic(options, value: bool) -> None:
  setattr(options, _get_option_deterministic_field(), value)


def _make_pathlike_fn(fn, nb_path_arg=1):
  """Wrap the function in a PathLike-compatible function."""

  @functools.wraps(fn)
  def new_fn(*args, **kwargs):
    # Normalize PathLike objects
    args = tuple(os.fspath(arg) for arg in args[:nb_path_arg])
    return fn(*args, **kwargs)

  return new_fn



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""Tool to generate the dataset catalog documentation.
"""

import argparse
import collections
import os
from typing import Dict, List, Optional

from absl import app
from absl.flags import argparse_flags
from etils import epath

import tensorflow_datasets as tfds
from tensorflow_datasets.scripts.documentation import doc_utils
from tensorflow_datasets.scripts.documentation import document_datasets
import yaml


def _parse_flags(_) -> argparse.Namespace:
  """Command line flags."""
  parser = argparse_flags.ArgumentParser(
      prog='build_catalog',
      description='Tool to generate the dataset catalog documentation',
  )
  parser.add_argument(
      '--datasets',
      help=(
          'Comma separated list of datasets to document. None for all datasets.'
      ),
  )
  parser.add_argument(
      '--ds_collections',
      help=(
          'Comma separated list of dataset collections to document. None for '
          'all datasets.'
      ),
  )
  parser.add_argument(
      '--catalog_dir',
      help='Directory path where to generate the catalog. Default to TFDS dir.',
  )
  return parser.parse_args()


def main(args: argparse.Namespace):
  # Note: The automated generation call `build_catalog.build_catalog()`
  # directly, so no code should go in `main()`.

  catalog_dir = args.catalog_dir or os.path.join(
      tfds.core.utils.tfds_write_path(),
      'docs',
      'catalog',
  )

  build_catalog(
      datasets=args.datasets.split(',') if args.datasets else None,
      ds_collections=(
          args.ds_collections.split(',') if args.ds_collections else None
      ),
      catalog_dir=catalog_dir,
  )


def _create_datasets_section_toc(
    section: str,
    builder_docs: List[document_datasets.BuilderDocumentation],
) -> str:
  """Creates the section of the overview.md table of content for datasets."""
  heading = '\n### `%s`\n' % section
  nightly_suffix = ' ' + doc_utils.NightlyDocUtil.icon
  entries = [
      f' * [`{doc.name}`]({doc.filestem}.md)'
      + (doc.is_nightly and nightly_suffix or '')
      for doc in builder_docs
  ]
  return '\n'.join([heading] + entries)


def _create_collections_section_toc(
    collection_docs: List[document_datasets.CollectionDocumentation],
) -> str:
  """Creates the section of the overview.md table of content for collections."""
  heading = '\n## `Dataset Collections`\n'
  entries = [f' * [`{doc.name}`]({doc.name}.md)' for doc in collection_docs]
  return '\n'.join([heading] + entries)


def build_catalog(
    datasets: Optional[List[str]] = None,
    ds_collections: Optional[List[str]] = None,
    *,
    catalog_dir: Optional[epath.PathLike] = None,
    doc_util_paths: Optional[doc_utils.DocUtilPaths] = None,
    toc_relative_path: str = '/datasets/catalog/',
    index_template: Optional[epath.PathLike] = None,
    index_filename: str = 'overview.md',
    dataset_types: Optional[List[tfds.core.visibility.DatasetType]] = None,
    include_collections: bool = True,
) -> None:
  """Document all datasets, including the table of content.

  Args:
    datasets: Lists of dataset to document (all if not set)
    ds_collections: Lists of all dataset collections to document (all if not
      set)
    catalog_dir: Destination path for the catalog
    doc_util_paths: Additional path for visualization, nightly info,...
    toc_relative_path: Relative path of the catalog directory, used to generate
      the table of content relative links.
    index_template: Default template for the index page.
    index_filename: Name of the catalog index file.
    dataset_types: Restrict the generation to the given dataset types. Default
      to all open source non-community datasets
    include_collections: Whether to include dataset collections to the catalog.
      Default to True.
  """
  dataset_types = dataset_types or [
      tfds.core.visibility.DatasetType.TFDS_PUBLIC,
  ]
  tfds.core.visibility.set_availables(dataset_types)

  catalog_dir = tfds.core.Path(catalog_dir)
  index_template = index_template or tfds.core.tfds_path(
      'scripts/documentation/templates/catalog_overview.md'
  )
  index_template = tfds.core.Path(index_template)

  # Iterate over the builder documentations
  section_to_builder_docs = collections.defaultdict(list)
  for builder_doc in document_datasets.iter_documentation_builders(
      datasets, doc_util_paths=doc_util_paths or doc_utils.DocUtilPaths()
  ):
    # Write the builder documentation
    dataset_file = catalog_dir / f'{builder_doc.filestem}.md'
    dataset_file.write_text(builder_doc.content)
    # Save the category
    for section in builder_doc.sections:
      section_to_builder_docs[section].append(builder_doc)

  section_to_collection_docs = collections.defaultdict(list)
  if include_collections:
    # Iterate over the dataset collection documentations
    for collection_doc in document_datasets.iter_collections_documentation(
        ds_collections
    ):
      # Write the dataset collection documentation
      collection_file = catalog_dir / f'{collection_doc.name}.md'
      collection_file.write_text(collection_doc.content)
      # Save the "dataset collection" docs
      section_to_collection_docs['collections'].append(collection_doc)

  _save_table_of_content(
      catalog_dir=catalog_dir,
      section_to_builder_docs=section_to_builder_docs,
      section_to_collection_docs=section_to_collection_docs,
      toc_relative_path=toc_relative_path,
      index_template=index_template,
      index_filename=index_filename,
      include_collections=include_collections,
  )


def _save_table_of_content(
    catalog_dir: tfds.core.Path,
    section_to_builder_docs: Dict[
        str, List[document_datasets.BuilderDocumentation]
    ],
    section_to_collection_docs: Dict[
        str, List[document_datasets.CollectionDocumentation]
    ],
    toc_relative_path: str,
    index_template: tfds.core.Path,
    index_filename: str,
    include_collections: bool,
) -> None:
  """Builds and saves the table of contents (`_toc.yaml` and `overview.md`)."""
  # For _toc.yaml
  toc_yaml = {
      'toc': [{
          'title': 'Overview',
          'path': os.path.join(toc_relative_path, 'overview'),
      }]
  }
  # For overview.md
  toc_overview = []

  # All collections documented, save the table of content
  if include_collections:
    for section, collection_docs in sorted(section_to_collection_docs.items()):
      collection_docs = sorted(collection_docs, key=lambda doc: doc.name)

      # Add `_toc.yaml` section
      sec_dict = {'title': 'Dataset Collections', 'section': []}
      for doc in collection_docs:
        sidebar_item = {
            'path': os.path.join(toc_relative_path, doc.name),
            'title': doc.name,
        }
        sec_dict['section'].append(sidebar_item)
      toc_yaml['toc'].append(sec_dict)

      # Add `overview.md` section
      toc_overview.append(_create_collections_section_toc(collection_docs))

  # All builder documented, save the table of content
  for section, builder_docs in sorted(section_to_builder_docs.items()):
    builder_docs = sorted(builder_docs, key=lambda doc: doc.name)

    # Add `_toc.yaml` section
    sec_dict = {'title': section, 'section': []}
    for doc in builder_docs:
      sidebar_item = {
          'path': os.path.join(toc_relative_path, doc.filestem),
          'title': doc.name + (' (manual)' if doc.is_manual else ''),
      }
      if doc.is_nightly:
        sidebar_item['status'] = 'nightly'
      sec_dict['section'].append(sidebar_item)
    toc_yaml['toc'].append(sec_dict)

    # Add `overview.md` section
    toc_overview.append(_create_datasets_section_toc(section, builder_docs))

  # Write the `overview.md` page
  index_str = index_template.read_text().format(toc='\n'.join(toc_overview))
  catalog_dir.joinpath(index_filename).write_text(index_str)

  # Write the `_toc.yaml` TF documentation navigation bar
  with catalog_dir.joinpath('_toc.yaml').open('w') as f:
    yaml.dump(toc_yaml, f, default_flow_style=False)


if __name__ == '__main__':
  app.run(main, flags_parser=_parse_flags)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""generates api_docs for tensorflow_datasets."""
import os

from absl import app
from absl import flags
from absl import logging

import tensorflow_datasets as tfds
# Testing is lazily imported, so we first force its import.
from tensorflow_datasets.testing import *  # pylint: disable=wildcard-import
from tensorflow_docs.api_generator import generate_lib

from tensorflow.tools.docs import doc_controls  # pylint: disable=g-direct-tensorflow-import

# Force the definition of all documentation decorators declared in
# third_party/py/tensorflow_datasets/core/utils/docs.py to TensorFlow decorators
try:
  tfds.core.utils.docs.deprecated = doc_controls.set_deprecated
  tfds.core.utils.docs.doc_private = doc_controls.doc_private
  tfds.core.utils.docs.do_not_doc = doc_controls.do_not_generate_docs
  # Same as `do_not_doc`, but also applied to children
  tfds.core.utils.docs.do_not_doc_inheritable = (
      doc_controls.do_not_doc_inheritable
  )
  # Document the parent, but not the children
  tfds.core.utils.docs.do_not_doc_in_subclasses = (
      doc_controls.do_not_doc_in_subclasses
  )
except AttributeError:
  logging.info("Could not set TensorFlow documentation decorators.")

FLAGS = flags.FLAGS

flags.DEFINE_string(
    "output_dir", "/tmp/datasets_api", "Where to output the docs"
)
flags.DEFINE_string(
    "code_url_prefix",
    "https://github.com/tensorflow/datasets/tree/master/tensorflow_datasets/",
    "The url prefix for links to code.",
)

flags.DEFINE_bool(
    "search_hints", True, "Include metadata search hints in the generated files"
)

flags.DEFINE_string(
    "site_path", "datasets/api_docs/python", "Path prefix in the _toc.yaml"
)


def execute(output_dir, code_url_prefix, search_hints, site_path):
  """Builds API docs for tensorflow_datasets."""
  doc_generator = generate_lib.DocGenerator(
      root_title="TensorFlow Datasets",
      py_modules=[("tfds", tfds)],
      base_dir=os.path.dirname(tfds.__file__),
      search_hints=search_hints,
      code_url_prefix=code_url_prefix,
      site_path=site_path,
  )

  doc_generator.build(output_dir)


def main(unused_argv):
  execute(
      FLAGS.output_dir,
      FLAGS.code_url_prefix,
      FLAGS.search_hints,
      FLAGS.site_path,
  )


if __name__ == "__main__":
  app.run(main)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""CNN/DailyMail Summarization dataset, non-anonymized version."""

import hashlib
import os
from typing import Dict, List

from absl import logging
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

_DESCRIPTION = """\
CNN/DailyMail non-anonymized summarization dataset.

There are two features:
  - article: text of news article, used as the document to be summarized
  - highlights: joined text of highlights with <s> and </s> around each
    highlight, which is the target summary
"""

# The second citation introduces the source data, while the first
# introduces the specific form (non-anonymized) we use here.
_CITATION = """\
@article{DBLP:journals/corr/SeeLM17,
  author    = {Abigail See and
               Peter J. Liu and
               Christopher D. Manning},
  title     = {Get To The Point: Summarization with Pointer-Generator Networks},
  journal   = {CoRR},
  volume    = {abs/1704.04368},
  year      = {2017},
  url       = {http://arxiv.org/abs/1704.04368},
  archivePrefix = {arXiv},
  eprint    = {1704.04368},
  timestamp = {Mon, 13 Aug 2018 16:46:08 +0200},
  biburl    = {https://dblp.org/rec/bib/journals/corr/SeeLM17},
  bibsource = {dblp computer science bibliography, https://dblp.org}
}

@inproceedings{hermann2015teaching,
  title={Teaching machines to read and comprehend},
  author={Hermann, Karl Moritz and Kocisky, Tomas and Grefenstette, Edward and Espeholt, Lasse and Kay, Will and Suleyman, Mustafa and Blunsom, Phil},
  booktitle={Advances in neural information processing systems},
  pages={1693--1701},
  year={2015}
}
"""

_DL_URLS = {
    # pylint: disable=line-too-long
    'cnn_stories': 'https://drive.google.com/uc?export=download&id=0BwmD_VLjROrfTHk4NFg2SndKcjQ',
    'dm_stories': 'https://drive.google.com/uc?export=download&id=0BwmD_VLjROrfM1BxdkxVaTY2bWs',
    'test_urls': 'https://raw.githubusercontent.com/abisee/cnn-dailymail/master/url_lists/all_test.txt',
    'train_urls': 'https://raw.githubusercontent.com/abisee/cnn-dailymail/master/url_lists/all_train.txt',
    'val_urls': 'https://raw.githubusercontent.com/abisee/cnn-dailymail/master/url_lists/all_val.txt',
    # pylint: enable=line-too-long
}

_HIGHLIGHTS = 'highlights'
_ARTICLE = 'article'
_PUBLISHER = 'publisher'
_ID = 'id'


def _get_url_hashes(path):
  """Get hashes of urls in file."""
  urls = _read_text_file(path)

  def url_hash(u):
    h = hashlib.sha1()
    try:
      u = u.encode('utf-8')
    except UnicodeDecodeError:
      logging.error('Cannot hash url: %s', u)
    h.update(u)
    return h.hexdigest()

  return {url_hash(u): True for u in urls}


def _find_files(dl_paths, publisher, url_dict):
  """Find files corresponding to urls."""
  if publisher == 'cnn':
    top_dir = os.path.join(dl_paths['cnn_stories'], 'cnn', 'stories')
  elif publisher == 'dm':
    top_dir = os.path.join(dl_paths['dm_stories'], 'dailymail', 'stories')
  else:
    logging.fatal('Unsupported publisher: %s', publisher)
  files = tf.io.gfile.listdir(top_dir)

  ret_files = []
  for p in files:
    basename = os.path.basename(p)
    if basename[0 : basename.find('.story')] in url_dict:
      ret_files.append(os.path.join(top_dir, p))
  return ret_files


def _subset_filenames(dl_paths, split):
  """Get filenames for a particular split."""
  assert isinstance(dl_paths, dict), dl_paths
  # Get filenames for a split.
  if split == tfds.Split.TRAIN:
    urls = _get_url_hashes(dl_paths['train_urls'])
  elif split == tfds.Split.VALIDATION:
    urls = _get_url_hashes(dl_paths['val_urls'])
  elif split == tfds.Split.TEST:
    urls = _get_url_hashes(dl_paths['test_urls'])
  else:
    logging.fatal('Unsupported split: %s', split)
  cnn = _find_files(dl_paths, 'cnn', urls)
  dm = _find_files(dl_paths, 'dm', urls)
  return {'cnn': cnn, 'dm': dm}


DM_SINGLE_CLOSE_QUOTE = '\u2019'  # unicode
DM_DOUBLE_CLOSE_QUOTE = '\u201d'
# acceptable ways to end a sentence
END_TOKENS = [
    '.',
    '!',
    '?',
    '...',
    "'",
    '`',
    '"',
    DM_SINGLE_CLOSE_QUOTE,
    DM_DOUBLE_CLOSE_QUOTE,
    ')',
]


def _read_text_file(text_file):
  lines = []
  with tf.io.gfile.GFile(text_file, 'r') as f:
    for line in f:
      lines.append(line.strip())
  return lines


def _get_art_abs(story_file):
  """Get abstract (highlights) and article from a story file path."""
  # Based on https://github.com/abisee/cnn-dailymail/blob/master/
  #     make_datafiles.py

  lines = _read_text_file(story_file)

  # The github code lowercase the text and we removed it in 3.0.0.

  # Put periods on the ends of lines that are missing them
  # (this is a problem in the dataset because many image captions don't end in
  # periods; consequently they end up in the body of the article as run-on
  # sentences)
  def fix_missing_period(line):
    """Adds a period to a line that is missing a period."""
    if '@highlight' in line:
      return line
    if not line:
      return line
    if line[-1] in END_TOKENS:
      return line
    return line + '.'

  lines = [fix_missing_period(line) for line in lines]

  # Separate out article and abstract sentences
  article_lines = []
  highlights = []
  next_is_highlight = False
  for line in lines:
    if not line:
      continue  # empty line
    elif line.startswith('@highlight'):
      next_is_highlight = True
    elif next_is_highlight:
      highlights.append(line)
    else:
      article_lines.append(line)

  # Make article into a single string
  article = ' '.join(article_lines)
  abstract = '\n'.join(highlights)

  return article, abstract


class CnnDailymail(tfds.core.GeneratorBasedBuilder):
  """CNN/DailyMail non-anonymized summarization dataset."""

  VERSION = tfds.core.Version('3.4.0')
  RELEASE_NOTES = {
      '1.0.0': 'New split API (https://tensorflow.org/datasets/splits)',
      '2.0.0': """
      Separate target sentences with newline. (Having the model predict newline
      separators makes it easier to evaluate using summary-level ROUGE.)
      """,
      '3.0.0': 'Using cased version.',
      '3.1.0': 'Removed BuilderConfig',
      '3.2.0': """
      Remove extra space before added sentence period.
      This shouldn't affect ROUGE scores because punctuation is removed.
      """,
      '3.3.0': 'Add publisher feature.',
      '3.4.0': 'Add ID feature.',
  }

  def _info(self):
    # Should return a tfds.core.DatasetInfo object
    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        features=tfds.features.FeaturesDict({
            _ARTICLE: tfds.features.Text(),
            _HIGHLIGHTS: tfds.features.Text(),
            _PUBLISHER: tfds.features.Text(),
            _ID: tfds.features.Text(),
        }),
        supervised_keys=(_ARTICLE, _HIGHLIGHTS),
        homepage='https://github.com/abisee/cnn-dailymail',
        citation=_CITATION,
    )

  def _split_generators(self, dl_manager):
    dl_paths = dl_manager.download_and_extract(_DL_URLS)

    return [
        tfds.core.SplitGenerator(
            name=tfds.Split.TRAIN,
            gen_kwargs={'files': _subset_filenames(dl_paths, tfds.Split.TRAIN)},
        ),
        tfds.core.SplitGenerator(
            name=tfds.Split.VALIDATION,
            gen_kwargs={
                'files': _subset_filenames(dl_paths, tfds.Split.VALIDATION)
            },
        ),
        tfds.core.SplitGenerator(
            name=tfds.Split.TEST,
            gen_kwargs={'files': _subset_filenames(dl_paths, tfds.Split.TEST)},
        ),
    ]

  def _generate_examples(self, files: Dict[str, List[str]]):
    for pub, file_list in files.items():
      for p in file_list:
        article, highlights = _get_art_abs(p)
        if not article or not highlights:
          continue
        fname = os.path.basename(p)
        clean_id = fname
        if fname.endswith('.story'):
          clean_id = clean_id[:-6]
        yield fname, {
            _ARTICLE: article,
            _HIGHLIGHTS: highlights,
            _PUBLISHER: pub,
            _ID: clean_id,
        }



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Wikipedia-based Image Text (WIT) Dataset."""

from __future__ import annotations

import csv
import os
import sys

from etils import epath
import numpy as np
import tensorflow_datasets.public_api as tfds

_DESCRIPTION = """
Wikipedia-based Image Text (WIT) Dataset is a large multimodal multilingual
dataset. WIT is composed of a curated set of 37.6 million entity rich image-text
examples with 11.5 million unique images across 108 Wikipedia languages. Its
size enables WIT to be used as a pretraining dataset for multimodal machine
learning models.
"""

_CITATION = """
@article{srinivasan2021wit,
  title={WIT: Wikipedia-based Image Text Dataset for Multimodal Multilingual Machine Learning},
  author={Srinivasan, Krishna and Raman, Karthik and Chen, Jiecao and Bendersky, Michael and Najork, Marc},
  journal={arXiv preprint arXiv:2103.01913},
  year={2021}
}
"""


class Wit(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for wit dataset."""

  VERSION = tfds.core.Version("1.1.0")
  RELEASE_NOTES = {
      "1.0.0": (
          "Initial release. It loads the WIT dataset from "
          "https://storage.googleapis.com/gresearch/wit/"
      ),
      "1.1.0": "Added `val` and `test` splits.",
  }

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        features=tfds.features.FeaturesDict({
            "language": tfds.features.Text(),
            "page_url": tfds.features.Text(),
            "image_url": tfds.features.Text(),
            "page_title": tfds.features.Text(),
            "section_title": tfds.features.Text(),
            "hierarchical_section_title": tfds.features.Text(),
            "caption_reference_description": tfds.features.Text(),
            "caption_attribution_description": tfds.features.Text(),
            "caption_alt_text_description": tfds.features.Text(),
            "mime_type": tfds.features.Text(),
            "original_height": np.int32,
            "original_width": np.int32,
            "is_main_image": np.bool_,
            "attribution_passes_lang_id": np.bool_,
            "page_changed_recently": np.bool_,
            "context_page_description": tfds.features.Text(),
            "context_section_description": tfds.features.Text(),
        }),
        supervised_keys=None,
        homepage="https://github.com/google-research-datasets/wit/",
        citation=_CITATION,
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""
    wit_homepage = "https://storage.googleapis.com/gresearch/wit/"
    wit_train_urls_to_download = [
        os.path.join(wit_homepage, f"wit_v1.train.all-0000{i}-of-00010.tsv.gz")
        for i in range(10)
    ]
    wit_val_urls_to_download = [
        os.path.join(wit_homepage, f"wit_v1.val.all-0000{i}-of-00005.tsv.gz")
        for i in range(5)
    ]
    wit_test_urls_to_download = [
        os.path.join(wit_homepage, f"wit_v1.test.all-0000{i}-of-00005.tsv.gz")
        for i in range(5)
    ]

    paths_per_split = dl_manager.download_and_extract({
        "train": wit_train_urls_to_download,
        "val": wit_val_urls_to_download,
        "test": wit_test_urls_to_download,
    })

    return {
        "train": self._generate_examples(paths_per_split["train"]),
        "val": self._generate_examples(paths_per_split["val"]),
        "test": self._generate_examples(paths_per_split["test"]),
    }

  def _generate_examples(self, filepaths):
    """Yields examples."""
    beam = tfds.core.lazy_imports.apache_beam

    def _process_example(elements):
      # Create a unique key for each example.
      filename, row_number, row = elements
      image_url = row["image_url"]
      example_key = f"{filename}_{row_number}_{image_url}"
      return example_key, {
          "language": row["language"],
          "page_url": row["page_url"],
          "image_url": row["image_url"],
          "page_title": row["page_title"],
          "section_title": row["section_title"],
          "hierarchical_section_title": row["hierarchical_section_title"],
          "caption_reference_description": row["caption_reference_description"],
          "caption_attribution_description": row[
              "caption_attribution_description"
          ],
          "caption_alt_text_description": row["caption_alt_text_description"],
          "mime_type": row["mime_type"],
          "original_height": int(row["original_height"]),
          "original_width": int(row["original_width"]),
          "is_main_image": bool(row["is_main_image"]),
          "attribution_passes_lang_id": bool(row["attribution_passes_lang_id"]),
          "page_changed_recently": bool(row["page_changed_recently"]),
          "context_page_description": row["context_page_description"],
          "context_section_description": (
              row["context_section_description"] or ""
          ),
      }

    def _read_rows(filename):
      # Limit to 100 MB. Value must be smaller than the C long maximum value.
      csv.field_size_limit(sys.maxsize)
      with epath.Path(filename).open() as f:
        csv_reader = csv.DictReader(
            f, delimiter="\t", quoting=csv.QUOTE_MINIMAL
        )
        for i, row in enumerate(csv_reader):
          yield filename, i, row

    return (
        beam.Create(filepaths)
        | beam.FlatMap(_read_rows)
        | beam.Map(_process_example)
    )



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Github API util tests."""

import contextlib
import os
import textwrap
from unittest import mock

from etils import epath
import pytest

from tensorflow_datasets.core.github_api import github_path

_SKIP_NON_HERMETIC = False

# Non hermetic tests are explicitly marked and skipped if `_SKIP_NON_HERMETIC`
# is True.
non_hermetic_test = pytest.mark.skipif(
    _SKIP_NON_HERMETIC,
    reason='Non-hermetic test skipped.',
)

_original_query_github = github_path.GithubApi.query

_AUTHOR_EXPECTED_CONTENT = textwrap.dedent(
    """\
    # This is the list of TensorFlow Datasets authors for copyright purposes.
    #
    # This does not necessarily list everyone who has contributed code, since in
    # some cases, their employer may be the copyright holder.  To see the full list
    # of contributors, see the revision history in source control.

    Google Inc.
    """
)

# Note: assert_no_api_call is globally applied on all tests (in conftest.py)


@contextlib.contextmanager
def enable_api_call():
  """Contextmanager which locally re-enable API calls."""
  with mock.patch.object(
      github_path.GithubApi, 'query', _original_query_github
  ):
    yield


def test_parse_github_path():
  url = 'github://tensorflow/datasets/tree/master/docs/README.md'
  repo, branch, path = github_path._parse_github_path(url)
  assert repo == 'tensorflow/datasets'
  assert branch == 'master'
  assert path == 'docs/README.md'

  url = 'github://tensorflow/datasets/tree/master'
  repo, branch, path = github_path._parse_github_path(url)
  assert repo == 'tensorflow/datasets'
  assert branch == 'master'
  assert path == ''  # pylint: disable=g-explicit-bool-comparison


def test_github_path_registered_as_path():
  uri = 'github://tensorflow/datasets/tree/master/docs/README.md'
  path = epath.Path(uri)
  assert isinstance(path, github_path.GithubPath)
  assert os.fspath(path) == uri


def test_invalid_github_path():
  # Path are lazily validated, so require explicit `_metadata` call.

  with pytest.raises(ValueError, match='Invalid github path'):
    _ = github_path.GithubPath()._metadata

  with pytest.raises(ValueError, match='Invalid github path'):
    _ = github_path.GithubPath('')._metadata

  with pytest.raises(ValueError, match='Invalid github path'):
    _ = github_path.GithubPath('github://not/a/path')

  with pytest.raises(ValueError, match='Invalid github path'):
    _ = github_path.GithubPath('github://tensorflow/tree/master/docs/README.md')

  # `blob` isn't accepted for consistency between paths.
  with pytest.raises(ValueError, match="/blob/` isn't accepted."):
    _ = github_path.GithubPath(
        'github://tensorflow/datasets/blob/master/docs/README.md'
    )

  p = github_path.GithubPath(
      'github://tensorflow/datasets/tree/master/docs/README.md'
  )
  p = p.parent  # /docs
  _ = p._metadata
  p = p.parent  # /
  _ = p._metadata
  p = p.parent
  with pytest.raises(ValueError, match='Invalid github path'):
    _ = p._metadata


def test_github_path_purepath():
  """Tests that pathlib methods works as expected."""
  p = github_path.GithubPath('github://tensorflow/datasets/tree/master')
  sub_p = p / 'some_folder'
  assert isinstance(sub_p, github_path.GithubPath)
  assert str(p) == 'github://tensorflow/datasets/tree/master'
  assert str(sub_p) == 'github://tensorflow/datasets/tree/master/some_folder'
  assert os.fspath(p) == 'github://tensorflow/datasets/tree/master'
  assert p == github_path.GithubPath.from_repo('tensorflow/datasets')


def test_github_path_as_url():
  p = github_path.GithubPath.from_repo('tensorflow/datasets', 'v3.1.0')
  p /= 'README.md'
  expected = (
      'https://raw.githubusercontent.com/tensorflow/datasets/v3.1.0/README.md'
  )
  assert p.as_raw_url() == expected


@non_hermetic_test
def test_github_api_listdir():
  """Test query github API."""
  # PurePath ops do not trigger API calls
  p = github_path.GithubPath.from_repo('tensorflow/datasets', 'v3.1.0')
  p = p / 'tensorflow_datasets' / 'testing'

  with enable_api_call():
    sub_dirs = sorted(p.iterdir())

  # `listdir` call cache the filetype of all childs
  all_dir_names = [d.name for d in sub_dirs if d.is_dir()]
  all_file_names = [d.name for d in sub_dirs if d.is_file()]
  all_names = [d.name for d in sub_dirs]

  with pytest.raises(NotADirectoryError):
    list((p / '__init__.py').iterdir())

  assert all_names == [
      '__init__.py',
      'dataset_builder_testing.py',
      'dataset_builder_testing_test.py',
      'fake_data_generation',
      'fake_data_utils.py',
      'generate_archives.sh',
      'metadata',
      'mocking.py',
      'mocking_test.py',
      'test_case.py',
      'test_data',
      'test_utils.py',
      'test_utils_test.py',
  ]
  assert all_dir_names == [
      'fake_data_generation',
      'metadata',
      'test_data',
  ]
  assert all_file_names == [
      '__init__.py',
      'dataset_builder_testing.py',
      'dataset_builder_testing_test.py',
      'fake_data_utils.py',
      'generate_archives.sh',
      'mocking.py',
      'mocking_test.py',
      'test_case.py',
      'test_utils.py',
      'test_utils_test.py',
  ]


@non_hermetic_test
def test_github_api_exists():
  """Test query github API."""
  p = github_path.GithubPath.from_repo('tensorflow/datasets', 'v3.1.0')
  with enable_api_call():
    assert p.exists()
    assert not (p / 'unknown_dir').exists()

  readme = p / 'README.md'
  core = p / 'tensorflow_datasets' / 'core'
  with enable_api_call():
    assert readme.is_file()
    assert core.is_dir()

  # Data should have been cached (no API calls required)
  assert not readme.is_dir()
  assert not core.is_file()
  assert readme.exists()
  assert core.exists()
  # Recreating a new Path reuse the cache
  readme_recreated = core.parent.parent / 'README.md'
  assert readme_recreated.is_file()
  assert readme_recreated._metadata == readme._metadata


@non_hermetic_test
def test_github_api_read_bytes_text():
  """Test query github API file content."""
  p = github_path.GithubPath.from_repo('tensorflow/datasets', 'v3.1.0')

  # Note: This is not wrapped inside `enable_api_call` contextmanager as
  # users need to download files without setting up an API token.

  content = (p / 'AUTHORS').read_bytes()
  assert isinstance(content, bytes)
  assert content == _AUTHOR_EXPECTED_CONTENT.encode()

  content = (p / 'AUTHORS').read_text()
  assert isinstance(content, str)
  assert content == _AUTHOR_EXPECTED_CONTENT

  # Cannot read the content of a directory.
  with pytest.raises(FileNotFoundError, match='Request failed'):
    (p / 'tensorflow_datasets' / 'core').read_bytes()


@non_hermetic_test
def test_github_api_copy(tmp_path):
  p = github_path.GithubPath.from_repo('tensorflow/datasets', 'v3.1.0')
  src = p / 'AUTHORS'
  dst = tmp_path / 'AUTHORS'

  target = src.copy(dst)
  assert target == dst
  assert dst.read_text() == _AUTHOR_EXPECTED_CONTENT

  with pytest.raises(FileExistsError, match='Destination .* exists'):
    src.copy(dst)

  src.copy(dst, overwrite=True)


def test_assert_no_api_call():
  with pytest.raises(AssertionError, match='Forbidden API call'):
    github_path.GithubPath.from_repo('tensorflow/datasets', 'v1.0.0').exists()


def test_get_tree():
  tree = {
      'tree': [
          {
              'path': 'code1.py',
              'type': 'blob',
          },
          {
              'path': 'myfolder',
              'type': 'tree',
          },
          {
              'path': 'myfolder/code2.py',
              'type': 'blob',
          },
          {
              'path': 'myfolder/mysubfolder',
              'type': 'tree',
          },
          {
              'path': 'myfolder/mysubfolder/code3.py',
              'type': 'blob',
          },
      ]
  }
  with mock.patch.object(github_path.GithubApi, 'query', return_value=tree):
    root = github_path.GithubPath.from_repo('tensorflow/datasets', 'v9.9.9')

    def gh_path(file: str) -> github_path.GithubPath:
      return github_path.GithubPath(
          f'github://tensorflow/datasets/tree/v9.9.9/{file}'
      )

    def assert_is_file(file):
      assert file.is_file()
      assert not file.is_dir()
      assert file.exists()

    def assert_is_folder(folder, files):
      assert set(folder.iterdir()) == files
      assert folder.is_dir()
      assert not folder.is_file()
      assert folder.exists()

    myfolder = gh_path('myfolder')
    mysubfolder = gh_path('myfolder/mysubfolder')
    code1 = gh_path('code1.py')
    code2 = gh_path('myfolder/code2.py')
    code3 = gh_path('myfolder/mysubfolder/code3.py')

    assert_is_folder(root, {code1, myfolder})
    assert_is_folder(myfolder, {code2, mysubfolder})
    assert_is_folder(mysubfolder, {code3})

    assert_is_file(code1)
    assert_is_file(code2)
    assert_is_file(code3)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""To associate metadata with TFDS calls."""

import enum
import threading
import time
from typing import Dict, Optional, Tuple

# Maps thread_id to "Session ID", if any.
_THREAD_TO_SESSIONID: Dict[int, int] = {}

_NEXT_SESSION_ID = 1
_NEXT_SESSION_ID_LOCK = threading.Lock()


class Status(enum.Enum):
  UNKNOWN = 0
  SUCCESS = 1
  ERROR = 2


def _get_session_id(thread_id: int) -> Tuple[int, bool]:
  """Returns (session_id, direct_call) tuple."""
  session_id = _THREAD_TO_SESSIONID.get(thread_id, None)
  if session_id:
    direct_call = False
  else:
    direct_call = True
    with _NEXT_SESSION_ID_LOCK:
      global _NEXT_SESSION_ID
      session_id = _NEXT_SESSION_ID
      _THREAD_TO_SESSIONID[thread_id] = session_id
      _NEXT_SESSION_ID += 1
  return session_id, direct_call


class CallMetadata:
  """Represents metadata associated with call.

  Object must be initialized just before the call, on same thread.
  """

  # The start and end times of the event (microseconds since Epoch).
  start_time_micros: Optional[int]
  end_time_micros: Optional[int]

  # The status (success or error) of the call.
  status: Status

  # Thread in which the call was made.
  thread_id: int

  # A unique number within the scope of the process to identify the session.
  # A session starts with a TFDS call (eg: `tfds.load(...)` and is propagated
  # through the stack to possibly other TFDS calls (eg: `tfds.dataset(...)`)
  # which the end-user did not call directly.
  session_id: int

  # Whether the operation was directly triggered by the user, ie: it is the
  # first operation of the session.
  direct_call: bool


  def __init__(self):
    self.status = Status.UNKNOWN
    self.thread_id = threading.get_ident()
    self.session_id, self.direct_call = _get_session_id(self.thread_id)
    self.start_time_micros = int(time.time() * 1e6)

  def mark_end(self):
    self.end_time_micros = int(time.time() * 1e6)
    if self.status == Status.UNKNOWN:
      self.status = Status.SUCCESS
    if self.direct_call:
      _THREAD_TO_SESSIONID.pop(self.thread_id)

  def mark_error(self):
    self.status = Status.ERROR



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utils for ConllUDatasetBuilder.

This file contains list of reusable tags which are commonly used in CoNLL-U
based datasets.
"""
import collections
from typing import Optional, OrderedDict

from tensorflow_datasets.core.dataset_builders.conll import conllu_dataset_builder
from tensorflow_datasets.core.features import feature as feature_lib
from tensorflow_datasets.core.features.class_label_feature import ClassLabel
from tensorflow_datasets.core.features.sequence_feature import Sequence
from tensorflow_datasets.core.features.text_feature import Text

# Universal part-of-speech tags.
UPOS = (
    "ADJ",
    "ADP",
    "ADV",
    "AUX",
    "CCONJ",
    "DET",
    "INTJ",
    "NOUN",
    "NUM",
    "PART",
    "PRON",
    "PROPN",
    "PUNCT",
    "SCONJ",
    "SYM",
    "VERB",
    "X",
    "_",
)

# Used by universal dependencies datasets.
UNIVERSAL_DEPENDENCIES_FEATURES = collections.OrderedDict({
    "idx": Text(),
    "tokens": Sequence(Text()),
    "lemmas": Sequence(Text()),
    "upos": Sequence(ClassLabel(names=UPOS)),
    "xpos": Sequence(Text()),
    "feats": Sequence(Text()),
    "head": Sequence(Text()),
    "deprel": Sequence(Text()),
    "deps": Sequence(Text()),
    "misc": Sequence(Text()),
    "text": Text(),
})

# Used for the xtreme's UD implementation.
XTREME_POS_FEATURES = collections.OrderedDict({
    "tokens": Sequence(Text()),
    "upos": Sequence(ClassLabel(names=UPOS)),
})


def get_universal_morphology_config(
    language: str,
    features: OrderedDict[str, feature_lib.FeatureConnector],
    name: Optional[str] = None,
    description: Optional[str] = None,
) -> conllu_dataset_builder.ConllUBuilderConfig:
  """Returns a populated Universal Morphology ConllUBuilderConfig.

  Args:
    language: The language of the data used to generate the ConllUBuilderConfig.
    features:  An OrderedDict specifying the features names and their type.
    name: Optional config name. If not given, will be set to be the same as the
      given language.
    description: Optional config description.

  Returns:
    A populated `ConllUBuilderConfig` for a Universal Morphology dataset.
  """
  return conllu_dataset_builder.ConllUBuilderConfig(
      name=name or language,
      description=description or None,
      features=features,
      language=language,
  )



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for conllu_dataset_builder."""
import textwrap
from unittest import mock

from etils import epath
import pytest
from tensorflow_datasets.core import lazy_imports_lib
from tensorflow_datasets.core.dataset_builders.conll import conllu_dataset_builder
from tensorflow_datasets.core.dataset_builders.conll import conllu_dataset_builder_utils as conllu_lib
import tensorflow_datasets.public_api as tfds

_FOLDER_PATH = "mock/path"

_VALID_INPUT = textwrap.dedent(
    """
# sent_id = VIT-9558
# text = Il futuro.
1	Il	il	DET	RD	Definite=Def|Gender=Masc|Number=Sing|PronType=Art	2	det	_	_
2	futuro	futuro	NOUN	S	Gender=Masc|Number=Sing	0	root	_	SpaceAfter=No
3	.	.	PUNCT	FS	_	2	punct	_	_

# sent_id = VIT-9478
# text = il responsabile dell'ambiente
1	il	il	DET	RD	Definite=Def|Gender=Masc|Number=Sing|PronType=Art	3	det	_	_
2	responsabile	responsabile	NOUN	S	Number=Sing	1	nsubj	_	_
3-4	dell'	_	_	_	_	_	_	_	SpaceAfter=No
3	di	di	ADP	E	_	6	case	_	_
4	l'	il	DET	RD	Definite=Def|Number=Sing|PronType=Art	6	det	_	_
5	ambiente	ambiente	NOUN	S	Gender=Masc|Number=Sing	3	nmod	_	_
"""
)

# The error making this invalid is the missing lemma field in line 1.
_INVALID_INPUT = textwrap.dedent(
    """
# sent_id = VIT-9558
# text = Il futuro.
1	Il	DET	RD	Definite=Def|Gender=Masc|Number=Sing|PronType=Art	2	det	_	_
2	futuro	futuro	NOUN	S	Gender=Masc|Number=Sing	0	root	_	SpaceAfter=No
3	.	.	PUNCT	FS	_	2	punct	_	_
"""
)

_INPUT_PATH = epath.Path(_FOLDER_PATH, "input_path.txt")


class DummyConllUDataset(conllu_dataset_builder.ConllUDatasetBuilder):
  VERSION = tfds.core.Version("1.0.0")
  RELEASE_NOTES = {"1.0.0": "Dummy notes."}
  BUILDER_CONFIGS = [
      conllu_lib.get_universal_morphology_config(
          language="italian",
          features=conllu_lib.UNIVERSAL_DEPENDENCIES_FEATURES,
      )
  ]

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    return self.create_dataset_info(
        description="Dummy CoNLL dataset.",
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""
    del dl_manager
    return {"train": self._generate_examples(_INPUT_PATH)}


def test_generate_example():
  tf_mock = mock.Mock()
  tf_mock.gfile.GFile.return_value = _VALID_INPUT
  expected_examples = []

  dataset = DummyConllUDataset()

  with tfds.testing.MockFs() as fs:
    fs.add_file(path=_INPUT_PATH, content=_VALID_INPUT)
    examples = list(dataset._generate_examples(_INPUT_PATH))
    expected_examples = [
        (
            0,
            {
                "idx": "VIT-9558",
                "text": "Il futuro.",
                "tokens": ["Il", "futuro", "."],
                "lemmas": ["il", "futuro", "."],
                "upos": ["DET", "NOUN", "PUNCT"],
                "xpos": ["RD", "S", "FS"],
                "feats": [
                    "{'Definite': 'Def', 'Gender': 'Masc', 'Number': 'Sing', "  # pylint:disable=implicit-str-concat
                    "'PronType': 'Art'}",
                    "{'Gender': 'Masc', 'Number': 'Sing'}",
                    "None",
                ],
                "head": ["2", "0", "2"],
                "deprel": ["det", "root", "punct"],
                "deps": ["None", "None", "None"],
                "misc": ["None", "{'SpaceAfter': 'No'}", "None"],
            },
        ),
        (
            1,
            {
                "idx": "VIT-9478",
                "text": "il responsabile dell'ambiente",
                "tokens": [
                    "il",
                    "responsabile",
                    "dell'",
                    "di",
                    "l'",
                    "ambiente",
                ],
                "lemmas": ["il", "responsabile", "_", "di", "il", "ambiente"],
                "upos": ["DET", "NOUN", "_", "ADP", "DET", "NOUN"],
                "xpos": ["RD", "S", "None", "E", "RD", "S"],
                "feats": [
                    "{'Definite': 'Def', 'Gender': 'Masc', 'Number': 'Sing', "  # pylint:disable=implicit-str-concat
                    "'PronType': 'Art'}",
                    "{'Number': 'Sing'}",
                    "None",
                    "None",
                    "{'Definite': 'Def', 'Number': 'Sing', 'PronType': 'Art'}",
                    "{'Gender': 'Masc', 'Number': 'Sing'}",
                ],
                "head": ["3", "1", "None", "6", "6", "3"],
                "deprel": ["det", "nsubj", "_", "case", "det", "nmod"],
                "deps": ["None", "None", "None", "None", "None", "None"],
                "misc": [
                    "None",
                    "None",
                    "{'SpaceAfter': 'No'}",
                    "None",
                    "None",
                    "None",
                ],
            },
        ),
    ]

    assert examples == expected_examples

    for _, example in examples:
      assert len(example) == len(conllu_lib.UNIVERSAL_DEPENDENCIES_FEATURES)

  assert len(examples) == 2


def test_generate_corrupted_example():
  conllu = lazy_imports_lib.lazy_imports.conllu

  tf_mock = mock.Mock()
  tf_mock.gfile.GFile.return_value = _VALID_INPUT
  dataset = DummyConllUDataset()

  with pytest.raises(conllu.exceptions.ParseException):
    with tfds.testing.MockFs() as fs:
      fs.add_file(path=_INPUT_PATH, content=_INVALID_INPUT)
      list(dataset._generate_examples(_INPUT_PATH))


class DummyXtremePosConllUDataset(conllu_dataset_builder.ConllUDatasetBuilder):
  VERSION = tfds.core.Version("1.0.0")
  RELEASE_NOTES = {"1.0.0": "Dummy notes."}
  BUILDER_CONFIGS = [
      conllu_lib.get_universal_morphology_config(
          language="italian", features=conllu_lib.XTREME_POS_FEATURES
      )
  ]

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    return self.create_dataset_info(
        description="Dummy CoNLL dataset.",
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""
    del dl_manager
    return {
        "train": self._generate_examples(
            filepaths=_INPUT_PATH,
            process_example_fn=conllu_dataset_builder.get_xtreme_pos_example,
        )
    }


def test_generate_xtreme_pos_example():
  tf_mock = mock.Mock()
  tf_mock.gfile.GFile.return_value = _VALID_INPUT
  expected_examples = []

  dataset = DummyXtremePosConllUDataset()

  with tfds.testing.MockFs() as fs:
    fs.add_file(path=_INPUT_PATH, content=_VALID_INPUT)
    examples = list(dataset._generate_examples(_INPUT_PATH))
    expected_examples = [
        (
            0,
            {
                "tokens": ["Il", "futuro", "."],
                "upos": ["DET", "NOUN", "PUNCT"],
            },
        ),
        (
            1,
            {
                "tokens": [
                    "il",
                    "responsabile",
                    "dell'",
                    "di",
                    "l'",
                    "ambiente",
                ],
                "upos": ["DET", "NOUN", "_", "ADP", "DET", "NOUN"],
            },
        ),
    ]

    assert examples == expected_examples

    for _, example in examples:
      assert len(example) == len(conllu_lib.XTREME_POS_FEATURES)

  assert len(examples) == 2


def test_generate_corrupted_xtreme_pos_example():
  conllu = lazy_imports_lib.lazy_imports.conllu

  tf_mock = mock.Mock()
  tf_mock.gfile.GFile.return_value = _VALID_INPUT
  dataset = DummyXtremePosConllUDataset()

  with pytest.raises(conllu.exceptions.ParseException):
    with tfds.testing.MockFs() as fs:
      fs.add_file(path=_INPUT_PATH, content=_INVALID_INPUT)
      list(dataset._generate_examples(_INPUT_PATH))



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base TestCase to use test_data."""

import contextlib
import os
import tempfile
from unittest import mock

from absl import logging
from tensorflow_datasets import setup_teardown
from tensorflow_datasets.core.utils import gcs_utils
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf

GCS_ACCESS_FNS = {
    "original_info": gcs_utils.gcs_dataset_info_files,
    "dummy_info": lambda _: [],
    "original_datasets": gcs_utils.is_dataset_on_gcs,
    "dummy_datasets": lambda _: False,
}


class TestCase(tf.test.TestCase):
  """Base TestCase to be used for all tests.

  `test_data` class attribute: path to the directory with test data.
  `tmp_dir` attribute: path to temp directory reset before every test.
  """

  # By default, tests globally applied fixture to disable GCS, Github API,...
  # If set, do not apply the given global fixtures
  DO_NOT_APPLY_FIXTURES = []

  @classmethod
  def setUpClass(cls):
    super(TestCase, cls).setUpClass()
    cls.test_data = os.path.join(os.path.dirname(__file__), "test_data")

    # TODO(tfds): Should make this a fixture.
    # Test must not communicate with GCS.
    gcs_utils.gcs_dataset_info_files = GCS_ACCESS_FNS["dummy_info"]
    gcs_utils.is_dataset_on_gcs = GCS_ACCESS_FNS["dummy_datasets"]

    # Apply the global fixtures as context managers
    cls._setup_cls_cms = []
    for fixture in setup_teardown.GLOBAL_FIXTURES:
      if fixture in cls.DO_NOT_APPLY_FIXTURES:
        continue
      cm = contextlib.contextmanager(fixture)()
      cm.__enter__()
      cls._setup_cls_cms.append(cm)

  @classmethod
  def tearDownClass(cls):
    for cm in cls._setup_cls_cms:
      cm.__exit__(None, None, None)

  @contextlib.contextmanager
  def gcs_access(self):
    # Restore GCS access
    gcs_utils.gcs_dataset_info_files = GCS_ACCESS_FNS["original_info"]
    gcs_utils.is_dataset_on_gcs = GCS_ACCESS_FNS["original_datasets"]
    yield
    # Revert access
    gcs_utils.gcs_dataset_info_files = GCS_ACCESS_FNS["dummy_info"]
    gcs_utils.is_dataset_on_gcs = GCS_ACCESS_FNS["dummy_datasets"]

  def setUp(self):
    super(TestCase, self).setUp()
    # get_temp_dir is actually the same for all tests, so create a temp sub-dir.
    self.tmp_dir = tempfile.mkdtemp(dir=tf.compat.v1.test.get_temp_dir())

  def assertRaisesWithPredicateMatch(self, err_type, predicate):
    if isinstance(predicate, str):
      predicate_fct = lambda err: predicate in str(err)
    else:
      predicate_fct = predicate
    return super(TestCase, self).assertRaisesWithPredicateMatch(
        err_type, predicate_fct
    )

  @contextlib.contextmanager
  def assertLogs(self, text, level="info"):
    with mock.patch.object(logging, level) as mock_log:
      yield
      concat_logs = ""
      for log_call in mock_log.call_args_list:
        args = log_call[0]
        base, args = args[0], args[1:]
        log_text = base % tuple(args)
        concat_logs += " " + log_text
      self.assertIn(text, concat_logs)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""DAVIS 2017 dataset for video object segmentation."""

from __future__ import annotations

import numpy as np
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

_DESCRIPTION = """
The DAVIS 2017 video object segmentation dataset.
"""

_URL = 'https://data.vision.ee.ethz.ch/csergi/share/davis/'

_CITATION = """\
@article{DBLP:journals/corr/Pont-TusetPCASG17,
  author    = {Jordi Pont{-}Tuset and
               Federico Perazzi and
               Sergi Caelles and
               Pablo Arbelaez and
               Alexander Sorkine{-}Hornung and
               Luc Van Gool},
  title     = {The 2017 {DAVIS} Challenge on Video Object Segmentation},
  journal   = {CoRR},
  volume    = {abs/1704.00675},
  year      = {2017},
  url       = {http://arxiv.org/abs/1704.00675},
  archivePrefix = {arXiv},
  eprint    = {1704.00675},
  timestamp = {Mon, 13 Aug 2018 16:48:55 +0200},
  biburl    = {https://dblp.org/rec/journals/corr/Pont-TusetPCASG17.bib},
  bibsource = {dblp computer science bibliography, https://dblp.org}
}
"""


def _read_annotation(annotation_path):
  """Read the image pixels of the annotation stored at annotations_path."""
  with tf.io.gfile.GFile(annotation_path, 'rb') as f:
    # Annotations are stored as index PNG so PIL returns directly the indices.
    image = tfds.core.lazy_imports.PIL_Image.open(f)
    image = np.asarray(image)
  return image[:, :, None]


class DavisConfig(tfds.core.BuilderConfig):
  """ "Configuration for RoboNet video rescaling."""

  def __init__(self, *, full_resolution=False, **kwargs):
    """The parameters specifying how the dataset will be processed.

    The dataset comes with two versions, 1080p and 480p video. Set
    full_resolution to True to load the 1080p version and to False to load the
    480p version.

    Args:
      full_resolution: bool, whether to load the 1080p version.
      **kwargs: Passed on to the constructor of `BuilderConfig`.
    """
    super(DavisConfig, self).__init__(**kwargs)
    self.full_resolution = full_resolution


class Davis(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for davis dataset."""

  BUILDER_CONFIGS = [
      DavisConfig(
          name='480p',
          description='The 480p version of the dataset',
          full_resolution=False,
      ),
      DavisConfig(
          name='full_resolution',
          description='The full resolution version of the dataset.',
          full_resolution=True,
      ),
  ]

  VERSION = tfds.core.Version('2.1.0')
  RELEASE_NOTES = {
      '1.0.0': 'Initial release.',
      '2.0.0': 'Change instance ids to be 0, 1, 2, ...',
      '2.1.0': 'Fix instance ids order.',
  }

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    # NOTE: Different videos have different resolutions.
    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        features=tfds.features.FeaturesDict({
            'video': tfds.features.Sequence({
                'frames': tfds.features.Image(shape=(None, None, 3)),
                'segmentations': tfds.features.Image(
                    shape=(None, None, 1),
                    use_colormap=True,
                ),
            }),
            'metadata': {
                'num_frames': np.int64,
                'video_name': np.str_,
            },
        }),
        supervised_keys=None,
        homepage='https://davischallenge.org/',
        citation=_CITATION,
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""

    if self.builder_config.full_resolution:
      trainval_data = dl_manager.download_and_extract(
          _URL + 'DAVIS-2017-trainval-Full-Resolution.zip'
      )
    else:
      trainval_data = dl_manager.download_and_extract(
          _URL + 'DAVIS-2017-trainval-480p.zip'
      )

    train_files = trainval_data / 'DAVIS/ImageSets/2017/train.txt'
    val_files = trainval_data / 'DAVIS/ImageSets/2017/val.txt'

    return {
        tfds.Split.TRAIN: self._generate_examples(train_files),
        tfds.Split.VALIDATION: self._generate_examples(val_files),
    }

  def _generate_examples(self, path):
    """Yields examples in the form of key, dataset dictionary."""

    videos_to_include = path.read_text().splitlines()
    root_path = path.parent.parent.parent  # Move up three directories.
    resolution = (
        'Full-Resolution' if self.builder_config.full_resolution else '480p'
    )
    for video in videos_to_include:
      images_path = root_path / 'JPEGImages' / resolution / video
      annotations_path = root_path / 'Annotations' / resolution / video
      seq_len = len(list(images_path.iterdir()))
      images = []
      annotations = []
      for i in range(seq_len):
        image_fname = f'{i:05d}.jpg'
        image_path = images_path / image_fname
        annotation_fname = f'{i:05d}.png'
        annotation_path = annotations_path / annotation_fname
        annotation = _read_annotation(annotation_path)
        images.append(image_path)
        annotations.append(annotation)

      video_dict = {'frames': images, 'segmentations': annotations}
      metadata = {'num_frames': seq_len, 'video_name': video}
      key = video + resolution
      yield key, {'video': video_dict, 'metadata': metadata}



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test for waymo_open_dataset."""

from tensorflow_datasets import testing
from tensorflow_datasets.object_detection import waymo_open_dataset


class WaymoOpenDatasetTest(testing.DatasetBuilderTestCase):
  """Waymo Open Dataset tests."""

  DATASET_CLASS = waymo_open_dataset.WaymoOpenDataset
  BUILDER_CONFIG_NAMES_TO_TEST = ["v1.0"]
  SPLITS = {
      "train": 1,  # Number of fake train example
      "validation": 1,  # Number of fake validation example
  }

  def setUp(self):
    """Set up Waymo Open Dataset tests."""
    super(WaymoOpenDatasetTest, self).setUp()
    for config in waymo_open_dataset.WaymoOpenDataset.BUILDER_CONFIGS:
      config.cloud_bucket = self.dummy_data


if __name__ == "__main__":
  testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for builder_templates."""
import pathlib
import pytest

from tensorflow_datasets.scripts.cli import builder_templates
from tensorflow_datasets.scripts.cli import cli_utils as utils

NAME = "my_dataset"
PATH_DIR = pathlib.Path.cwd() / "tensorflow_datasets/image/my_dataset"
TFDS_API = "tensorflow_datasets"
TODO = f"{NAME}"

testdata = [
    (builder_templates.STANDARD, "tfds.core.GeneratorBasedBuilder"),
    (builder_templates.CONLL, "tfds.dataset_builders.ConllDatasetBuilder"),
    (builder_templates.CONLLU, "tfds.dataset_builders.ConllUDatasetBuilder"),
]


@pytest.mark.parametrize("ds_format,ds_builder", testdata)
def test_valid_builder_templates(ds_format, ds_builder):
  dataset_info = utils.DatasetInfo(
      name=NAME, in_tfds=True, path=PATH_DIR, data_format=ds_format
  )
  template = builder_templates.create_builder_template(dataset_info)
  assert isinstance(template, str)
  assert ds_builder in template


def test_create_inexistent_builder_template():
  nonexistent_data_format = "nonexistent_data_format"
  nonexistent_dataset_info = utils.DatasetInfo(
      name=NAME,
      in_tfds=True,
      path=PATH_DIR,
      data_format=nonexistent_data_format,
  )

  error_msg = (
      f"Required format {nonexistent_data_format} isn't associated with a"
      " format-specific builder in TFDS."
  )
  with pytest.raises(ValueError, match=error_msg):
    builder_templates.create_builder_template(nonexistent_dataset_info)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""tensorflow/datasets is a library of datasets ready to use with TensorFlow.

tensorflow/datasets is a library of public datasets ready to use with
TensorFlow. Each dataset definition contains the logic necessary to download and
prepare the dataset, as well as to read it into a model using the
`tf.data.Dataset` API.

Usage outside of TensorFlow is also supported.

See the README on GitHub for further documentation.
"""

import datetime
import itertools
import os
import sys

import pkg_resources
from setuptools import find_packages
from setuptools import setup

if '--nightly' in sys.argv:
  nightly = True
  sys.argv.remove('--nightly')
else:
  nightly = False

project_name = 'tensorflow-datasets'

# To enable importing version.py directly, we add its path to sys.path.
version_path = os.path.join(os.path.dirname(__file__), 'tensorflow_datasets')
sys.path.append(version_path)
from version import __version__  # pytype: disable=import-error  # pylint: disable=g-import-not-at-top

if nightly:
  project_name = 'tfds-nightly'
  # Version as `X.Y.Z.dev199912312459`
  datestring = os.environ.get(
      'TFDS_NIGHTLY_TIMESTAMP'
  ) or datetime.datetime.now().strftime('%Y%m%d%H%M')
  curr_version = pkg_resources.parse_version(__version__)
  __version__ = f'{curr_version.base_version}.dev{datestring}'

DOCLINES = __doc__.split('\n')

REQUIRED_PKGS = [
    'absl-py',
    'array_record',
    'click',
    'dm-tree',
    'etils[enp,epath,etree]>=0.9.0',
    'numpy',
    'promise',
    'protobuf>=3.20',
    'psutil',
    'requests>=2.19.0',
    'tensorflow-metadata',
    'termcolor',
    'toml',
    'tqdm',
    'wrapt',
    # Standard library backports
    'importlib_resources;python_version<"3.9"',
]

TESTS_DEPENDENCIES = [
    'dill',
    'jax[cpu]',
    'jupyter',
    'pytest',
    'pytest-shard',
    'pytest-xdist',
    # Lazy-deps required by core
    'pandas',
    'pydub',
    'apache-beam',
    'conllu',
    # TODO(b/142892342): Re-enable
    # 'tensorflow-docs @ git+https://github.com/tensorflow/docs#egg=tensorflow-docs',  # pylint: disable=line-too-long
    # Required by scripts/documentation/
    'pyyaml',
    'tensorflow-io[tensorflow]',
    # `datasets` needs to be installed separately in Python >= 3.10 due to
    # conflicts between `multiprocess` and `apache-beam` libraries. See
    # https://github.com/uqfoundation/multiprocess/issues/125
    'datasets;python_version<"3.10"',
]

# Additional deps for formatting
DEV_DEPENDENCIES = [
    'pylint>=2.6.0',
    'yapf',
]

# Static files needed by datasets.
DATASET_FILES = [
    'datasets/imagenet2012/labels.txt',
    'datasets/imagenet2012/validation_labels.txt',
    'datasets/lvis/classes.txt',
    'datasets/ogbg_molpcba/tasks.txt',
    'datasets/quickdraw/labels.txt',
    'datasets/smartwatch_gestures/labels.txt',
    'image_classification/caltech101_labels.txt',
    'image_classification/categories_places365.txt',
    'image_classification/cbis_ddsm_calc_distributions.txt',
    'image_classification/cbis_ddsm_calc_types.txt',
    'image_classification/cbis_ddsm_mass_margins.txt',
    'image_classification/cbis_ddsm_mass_shapes.txt',
    'image_classification/cbis_ddsm_patch_labels.txt',
    'image_classification/dtd_key_attributes.txt',
    'image_classification/food-101_classes.txt',
    'image_classification/i_naturalist2018/inaturalist2018_labels.txt',
    'image_classification/i_naturalist2018/inaturalist2018_supercategories.txt',
    'image_classification/i_naturalist2021/i_naturalist2021_labels.txt',
    'image_classification/i_naturalist2021/i_naturalist2021_supercategories.txt',
    'image_classification/imagenet_resized_labels.txt',
    'image_classification/imagenette_labels.txt',
    'image_classification/imagewang_labels.txt',
    'image_classification/inaturalist_labels.txt',
    'image_classification/inaturalist_supercategories.txt',
    'image_classification/placesfull/categories_placesfull.txt',
    'image_classification/plantae_k_urls.txt',
    'image_classification/sun397_labels.txt',
    'image_classification/sun397_tfds_te.txt',
    'image_classification/sun397_tfds_tr.txt',
    'image_classification/sun397_tfds_va.txt',
    'object_detection/open_images_classes_all.txt',
    'object_detection/open_images_classes_boxable.txt',
    'object_detection/open_images_classes_trainable.txt',
    'video/tao/labels.txt',
    'video/ucf101_labels.txt',
    'video/youtube_vis/labels.txt',
]

# Extra dependencies required by specific datasets
DATASET_EXTRAS = {
    # In alphabetical order
    'aflw2k3d': ['scipy'],
    'beir': ['apache-beam'],
    'ble_wind_field': ['gcsfs', 'zarr'],
    'c4': ['apache-beam', 'gcld3', 'langdetect', 'nltk', 'tldextract'],
    'c4_wsrs': ['apache-beam'],
    'cats_vs_dogs': ['matplotlib'],
    'colorectal_histology': ['Pillow'],
    'common_voice': ['pydub'],  # and ffmpeg installed
    'duke_ultrasound': ['scipy'],
    'eurosat': ['scikit-image', 'tifffile', 'imagecodecs'],
    'groove': ['pretty_midi', 'pydub'],
    'gtzan': ['pydub'],
    'imagenet2012_corrupted': [
        # This includes pre-built source; you may need to use an alternative
        # route to install OpenCV
        'opencv-python',
        'scikit-image',
        'scipy',
    ],
    'librispeech': ['pydub'],  # and ffmpeg installed
    'lsun': ['tensorflow-io[tensorflow]'],
    # sklearn version required to avoid conflict with librosa from
    # https://github.com/scikit-learn/scikit-learn/issues/14485
    # See https://github.com/librosa/librosa/issues/1160
    'nsynth': ['crepe>=0.0.11', 'librosa', 'scikit-learn==0.20.3'],
    'ogbg_molpcba': ['pandas', 'networkx'],
    'pet_finder': ['pandas'],
    'robonet': ['h5py'],  # and ffmpeg installed
    # envlogger is not available for Python versions >= 3.10 or non Linux
    # platforms: https://pypi.org/project/envlogger/#files
    # tests are disabled in `tensorflow_datasets/conftest.py`
    'locomotion': ['envlogger;python_version<"3.10" and sys_platform=="linux"'],
    'robosuite_panda_pick_place_can': [
        'envlogger;python_version<"3.10" and sys_platform=="linux"'
    ],
    'smartwatch_gestures': ['pandas'],
    'svhn': ['scipy'],
    'the300w_lp': ['scipy'],
    'wider_face': ['Pillow'],
    'wiki_dialog': ['apache-beam'],
    'wikipedia': ['apache-beam', 'mwparserfromhell', 'mwxml'],
    'wsc273': ['bs4', 'lxml'],
    'youtube_vis': ['pycocotools'],
}

# Those datasets have dependencies which conflict with the rest of TFDS, so
# running them in an isolated environments.
ISOLATED_DATASETS = ('nsynth', 'lsun')

# Extra dataset deps are required for the tests
all_dataset_dependencies = list(
    itertools.chain.from_iterable(
        deps
        for ds_name, deps in DATASET_EXTRAS.items()
        if ds_name not in ISOLATED_DATASETS
    )
)

TESTS_ALL_DEPENDENCIES = TESTS_DEPENDENCIES + all_dataset_dependencies
HUGGINGFACE_ALL_DEPENDENCIES = [
    dep
    for dep in TESTS_ALL_DEPENDENCIES
    if not dep.startswith('apache-beam') and not dep.startswith('datasets')
] + ['datasets']

EXTRAS = {
    'matplotlib': ['matplotlib'],
    'tensorflow': ['tensorflow>=2.1'],
    'tf-nightly': ['tf-nightly'],
    'tensorflow-data-validation': ['tensorflow-data-validation'],
    'tests-all': TESTS_ALL_DEPENDENCIES,
    'dev': TESTS_DEPENDENCIES + DEV_DEPENDENCIES,
    'huggingface': HUGGINGFACE_ALL_DEPENDENCIES,
}
EXTRAS.update(DATASET_EXTRAS)

setup(
    name=project_name,
    version=__version__,
    description=DOCLINES[0],
    long_description='\n'.join(DOCLINES[2:]),
    author='Google Inc.',
    author_email='packages@tensorflow.org',
    url='https://github.com/tensorflow/datasets',
    download_url='https://github.com/tensorflow/datasets/tags',
    license='Apache 2.0',
    packages=find_packages(),
    package_data={
        'tensorflow_datasets': DATASET_FILES + [
            # Bundle `datasets/` folder in PyPI releases
            'datasets/*/*',
            'core/utils/colormap.csv',
            'scripts/documentation/templates/*',
            'url_checksums/*',
            'checksums.tsv',
            'community-datasets.toml',
            'dataset_collections/*/*.md',
            'dataset_collections/*/*.bib',
            'core/valid_tags.txt',
        ],
    },
    exclude_package_data={
        'tensorflow_datasets': [
            'dummy_data/*',
        ],
    },
    scripts=[],
    install_requires=REQUIRED_PKGS,
    python_requires='>=3.9',
    extras_require=EXTRAS,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
    ],
    keywords='tensorflow machine learning datasets',
    entry_points={
        'console_scripts': [
            'tfds = tensorflow_datasets.scripts.cli.main:launch_cli'
        ],
    },
    # Include_package_data is required for setup.py to recognize the MANIFEST.in
    #   https://python-packaging.readthedocs.io/en/latest/non-code-files.html
    include_package_data=True,
)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Builder for DMLab Datasets."""

from __future__ import annotations

import dataclasses
from typing import Any, Dict

import numpy as np
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds
from tensorflow_datasets.rl_unplugged import rlu_common

_DMLAB_DESCRIPTION = """
DeepMind Lab dataset has several levels from the challenging, partially
observable [Deepmind Lab suite](https://github.com/deepmind/lab). DeepMind Lab
dataset is collected by training distributed R2D2 by [Kapturowski et al., 2018]
(https://openreview.net/forum?id=r1lyTjAqYX) agents from scratch on individual
tasks. We recorded the experience across all actors during entire training runs
a few times for every task. The details of the dataset generation process is
described in [Gulcehre et al., 2021](https://arxiv.org/abs/2103.09575).

We release datasets for five different DeepMind Lab levels: `seekavoid_arena_01`,
`explore_rewards_few`, `explore_rewards_many`, `rooms_watermaze`,
`rooms_select_nonmatching_object`. We also release the snapshot datasets for
`seekavoid_arena_01` level that we generated the datasets from a trained R2D2
snapshot with different levels of epsilons for the epsilon-greedy algorithm
when evaluating the agent in the environment.

DeepMind Lab dataset is fairly large-scale. We recommend you to try it if you
are interested in large-scale offline RL models with memory.
"""

_CITATION = """
@article{gulcehre2021rbve,
    title={Regularized Behavior Value Estimation},
    author={{\\c{C}}aglar G{\\"{u}}l{\\c{c}}ehre and
               Sergio G{\\'{o}}mez Colmenarejo and
               Ziyu Wang and
               Jakub Sygnowski and
               Thomas Paine and
               Konrad Zolna and
               Yutian Chen and
               Matthew W. Hoffman and
               Razvan Pascanu and
               Nando de Freitas},
    year={2021},
    journal   = {CoRR},
    url       = {https://arxiv.org/abs/2103.09575},
    eprint={2103.09575},
    archivePrefix={arXiv},
}
"""

_PIXELS_HEIGHT = 72
_PIXELS_WIDTH = 96
_PIXELS_SHAPE = (_PIXELS_HEIGHT, _PIXELS_WIDTH, 3)


@dataclasses.dataclass
class BuilderConfig(tfds.core.BuilderConfig):
  """Configuration of the task.

  Attributes:
    task: name of the DMLab task
    episode_length: length of the episodes in this task
  """

  task: str = 'seekavoid_arena01'
  episode_length: int = 301


class DMLabDatasetBuilder(rlu_common.RLUBuilder, skip_registration=True):
  """DatasetBuilder for RLU DMLab."""

  _SHARDS = 500
  _INPUT_FILE_PREFIX = 'gs://rl_unplugged/dmlab/'

  def get_features_dict(self):
    return tfds.features.FeaturesDict({
        'steps': tfds.features.Dataset({
            'observation': {
                'pixels': tfds.features.Image(
                    shape=_PIXELS_SHAPE, dtype=np.uint8, encoding_format='png'
                ),
                'last_action': np.int64,
                'last_reward': np.float32,
            },
            'action': np.int64,
            'reward': np.float32,
            'is_terminal': np.bool_,
            'is_first': np.bool_,
            'is_last': np.bool_,
            'discount': np.float32,
        }),
        'episode_id': np.int64,
        'episode_return': np.float32,
    })

  def get_description(self):
    return _DMLAB_DESCRIPTION

  def get_citation(self):
    return _CITATION

  def get_file_prefix(self):
    run = self.builder_config.name
    task = self.builder_config.task
    return f'{self._INPUT_FILE_PREFIX}/{task}/{run}/tfrecord'

  def num_shards(self):
    return self._SHARDS

  def tf_example_to_step_ds(
      self, tf_example: tf.train.Example
  ) -> Dict[str, Any]:
    """Create an episode from a TF example."""
    episode_length = self.builder_config.episode_length

    # Parse tf.Example.
    def sequence_feature(shape, dtype=tf.float32):
      return tf.io.FixedLenFeature(shape=[episode_length] + shape, dtype=dtype)

    feature_description = {
        'episode_id': tf.io.FixedLenFeature([], tf.int64),
        'start_idx': tf.io.FixedLenFeature([], tf.int64),
        'episode_return': tf.io.FixedLenFeature([], tf.float32),
        'observations_pixels': sequence_feature([], tf.string),
        'observations_reward': sequence_feature([]),
        # actions are one-hot arrays.
        'observations_action': sequence_feature([15]),
        'actions': sequence_feature([], tf.int64),
        'rewards': sequence_feature([]),
        'discounted_rewards': sequence_feature([]),
        'discounts': sequence_feature([]),
    }

    data = tf.io.parse_single_example(tf_example, feature_description)

    # Dmlab observations are compressed with OpenCV imencode function. Encoding
    # with imencode and decoding with TF PNG decoder flips the order of the
    # color channels. Here we flip back the order color of channels from BGR to
    # RGB:
    pixels = tf.scan(
        fn=lambda _, png: tf.reshape(tf.io.decode_png(png), _PIXELS_SHAPE),
        elems=data['observations_pixels'],
        initializer=tf.zeros(_PIXELS_SHAPE, dtype=tf.uint8),
    )
    pixels = tf.reverse(pixels, axis=[-1])

    episode = {
        # Episode Metadata
        'episode_id': data['episode_id'],
        'episode_return': data['episode_return'],
        'steps': {
            'observation': {
                'pixels': pixels,
                'last_action': tf.argmax(
                    data['observations_action'], axis=1, output_type=tf.int64
                ),
                'last_reward': data['observations_reward'],
            },
            'action': data['actions'],
            'reward': data['rewards'],
            'discount': data['discounts'],
            'is_first': [True] + [False] * (episode_length - 1),
            'is_last': [False] * (episode_length - 1) + [True],
            'is_terminal': [False] * (episode_length),
        },
    }
    return episode
